SystemOrganization addCategory: #'Ma special collections'!
SystemOrganization addCategory: #'Ma special collections-Info'!
SystemOrganization addCategory: #'Ma special collections-Numbers collections'!
SystemOrganization addCategory: #'Ma special collections-Squeak weak tweaks'!
SystemOrganization addCategory: #'Ma special collections-Tree'!
SystemOrganization addCategory: #'Ma special collections-private'!
SystemOrganization addCategory: #'Ma special collections-Exceptions'!
SystemOrganization addCategory: #'Ma special collections-Records'!
SystemOrganization addCategory: #'Ma special collections-exceptions'!

MaObject subclass: #MaDictionary
	instanceVariableNames: 'dicts'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Squeak weak tweaks'!

!MaDictionary commentStamp: 'cmm 12/28/2005 12:50' prior: 0!
I am identical to a Dictionary but a lot faster when my size grows beyond 4K elements.  Squeak uses a 12-bit identity hash for its object identity, causing lots of hash collisions once you get more than 4K (2 raisedTo: 12) objects in it.  As this script demonstrates, I am comparable at adding, accessing and replacing, but much faster with removing.

| sd md r |
Transcript clear.
r _ (1 to: 15500) collect: [ :e | Object new ].
r _ r shuffled.
sd _ WeakIdentityKeyDictionary new.
md _ MaWeakIdentityKeyDictionary new.
Transcript cr; show: 'time to add to sd: ', ([ r do: [ :each | sd at: each put: each ] ] timeToRun) printString.
Transcript cr; show: 'time to add to md: ', ([ r do: [ :each | md at: each put: each ] ] timeToRun) printString.
r_r shuffled.
Transcript cr; show: 'time to access to sd: ', ([ r do: [ :each | sd at: each ] ] timeToRun) printString.
Transcript cr; show: 'time to access to md: ', ([ r do: [ :each | md at: each ] ] timeToRun) printString.
r _ r shuffled.
Transcript cr; show: 'time to replace to sd: ', ([ r do: [ :each | sd at: each put: each ] ] timeToRun) printString.
Transcript cr; show: 'time to replace to md: ', ([ r do: [ :each | md at: each put: each ] ] timeToRun) printString.
r _ r shuffled.
Transcript cr; show: 'time to remove 300 from sd: ', ([ (r copyFrom: 1 to: 300) do: [ :each | sd removeKey: each ] ] timeToRun) printString.
Transcript cr; show: 'time to remove 300 from md: ', ([ (r copyFrom: 1 to: 300) do: [ :each | md removeKey: each ] ] timeToRun) printString.
!

!MaDictionary class methodsFor: 'creation' stamp: 'cmm 12/27/2002 10:09'!
new

	^self new: 8! !

!MaDictionary class methodsFor: 'creation' stamp: 'cmm 12/13/2002 14:28'!
new: anInteger

	^super new capacity: anInteger; yourself! !

!MaDictionary methodsFor: 'adding' stamp: 'cmm 12/12/2002 17:10'!
add: anAssociation

	self
		removeKey: anAssociation key
		ifAbsent: [ "do nothing" ].
	self qualifyingDict add: anAssociation! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/22/2002 18:26'!
associationsDo: aBlock

	dicts do: [ :each | each associationsDo: aBlock ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 16:43'!
at: anObject

	^self
		at: anObject
		ifAbsent: [ self errorKeyNotFound ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 16:44'!
at: key ifAbsentPut: aBlock

	^self
		at: key
		ifAbsent:
			[ self 
				at: key 
				put: aBlock value ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/13/2002 14:34'!
at: anObject ifAbsent: aBlock

	| uniqueObject |
	uniqueObject _ ''.
	dicts do:
		[ :each |
		| result |
		result _ each
			at: anObject
			ifAbsent: [ uniqueObject ].
		result == uniqueObject ifFalse: [ ^result ] ].
	^aBlock value! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 6/8/2004 12:51'!
at: key put: anObject

	key ifNil: [ self error: 'nil cannot be used as a key to a Dictionary.' ].
	dicts do:
		[ :each |
		(each includesKey: key) ifTrue: [ ^each at: key put: anObject ] ].
	^self qualifyingDict
		at: key
		put: anObject! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/27/2002 09:57'!
capacity: anInteger

	| numberOfDicts |
	numberOfDicts _ anInteger // self criticalSize.
	dicts _ OrderedCollection new: numberOfDicts.
	numberOfDicts timesRepeat: [ dicts add: (self squeakDictionaryClass new: self criticalSize) ].
	anInteger \\ self criticalSize > 0
		ifTrue:
			[ dicts add: (self squeakDictionaryClass new: anInteger \\ self criticalSize) ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 17:17'!
collect: aBlock 
	"Evaluate aBlock with each of my values as the argument.  Collect the
	resulting values into a collection that is like me. Answer with the new
	collection."
	| newCollection |
	newCollection _ OrderedCollection new: self size.
	self do: [:each | newCollection add: (aBlock value: each)].
	^ newCollection! !

!MaDictionary methodsFor: 'copying' stamp: 'cmm 3/24/2005 12:04'!
copy

	^ self copy postCopy! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/27/2002 10:05'!
criticalSize
	"Since the Squeak Set allocates about 4/3rd's extra space for future growth, we have the potential to grow a little out of its efficiency range.  Compensate."

	^4096*3//4! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 2/5/2003 20:47'!
detect: aBlock

	^self
		detect: aBlock
		ifNone: [ self error: 'not in MaDictionary' ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 2/5/2003 20:47'!
detect: aBlock ifNone: notFoundBlock

	self do: [ :each | (aBlock value: each) ifTrue: [ ^each ] ].
	^notFoundBlock value! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 16:58'!
do: aBlock

	dicts do: [ :each | each do: aBlock ]! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 16:43'!
errorKeyNotFound

	self error: 'key not found'! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/13/2002 14:03'!
finishCopy

	dicts _ dicts collect: [ :each | each copy ]! !

!MaDictionary methodsFor: 'testing' stamp: 'cmm 12/12/2002 17:14'!
includesKey: key 
	"Answer whether the receiver has a key equal to the argument, key."
	
	self at: key ifAbsent: [^false].
	^true! !

!MaDictionary methodsFor: 'testing' stamp: 'cmm 12/12/2002 17:14'!
includes: anObject

	self do: [:each | anObject = each ifTrue: [^true]].
	^false! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 16:52'!
keys
	"Answer a Set containing the receiver's keys."
	| aSet |
	aSet _ Set new: self size.
	self keysDo: [:key | aSet add: key].
	^ aSet! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/25/2002 23:28'!
keysAndValuesDo: twoArgBlock

	dicts do: [ :each | each keysAndValuesDo: twoArgBlock ]! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:06'!
keysAndValuesRemove: twoArgBlock

	dicts do: [ :each | each keysAndValuesRemove: twoArgBlock ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 16:52'!
keysDo: aBlock

	dicts do: [ :each | each keysDo: aBlock ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 17:12'!
maAllKeysAtValue: anObject

	| aCollection |
	aCollection := OrderedCollection new.
	self keysAndValuesDo: [ :key :value | value = anObject ifTrue: [ aCollection add: key ] ].
	^aCollection! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/13/2002 14:33'!
maAt: anObject ifPresent: oneArgBlock ifAbsent: zeroArgBlock

	| current uniqueObject |

	uniqueObject _ ''.
	^(current := self
		at: anObject
		ifAbsent: [ uniqueObject ]) == uniqueObject
			ifTrue: [ zeroArgBlock value ]
			ifFalse: [ oneArgBlock value: current ]! !

!MaDictionary methodsFor: 'testing' stamp: 'cmm 12/12/2002 17:15'!
maIsHashedCollection

	^true! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 17:16'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject."

	| count |
	count _ 0.
	self do: [:each | anObject = each ifTrue: [count _ count + 1]].
	^count! !

!MaDictionary methodsFor: 'copying' stamp: 'cmm 5/2/2005 22:13'!
postCopy
	super postCopy.
	dicts := dicts copy.
	self finishCopy! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 23:20'!
qualifyingDict

	^dicts
		detect: [ :each | each size < self criticalSize ]
		ifNone: [ dicts addFirst: self squeakDictionaryClass new ]! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 17:15'!
rehash

	dicts do: [ :each | each rehash ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 1/9/2003 00:40'!
reject: aBlock 
	"Evaluate aBlock with each of my values as the argument. Collect into a
	new dictionary, only those associations for which aBlock evaluates to
	true."

	| newCollection |
	newCollection _ self species new.
	self associationsDo: 
		[:each | 
		(aBlock value: each value) ifFalse: [newCollection add: each]].
	^newCollection! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:05'!
removeKey: key 
	"Remove key from the receiver.
	If key is not in the receiver, notify an error."

	^ self removeKey: key ifAbsent: [self errorKeyNotFound]! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/13/2002 14:34'!
removeKey: key ifAbsent: aBlock

	| uniqueObject |
	uniqueObject _ ''.
	dicts do:
		[ :each |
		| removed |
		removed _ each
			removeKey: key
			ifAbsent: [ uniqueObject ].
		removed == uniqueObject ifFalse: [ ^removed ] ].
	^aBlock value! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:04'!
remove: anObject

	self shouldNotImplement! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:04'!
remove: anObject ifAbsent: exceptionBlock

	self shouldNotImplement! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 17:09'!
select: aBlock 
	"Evaluate aBlock with each of my values as the argument. Collect into a
	new dictionary, only those associations for which aBlock evaluates to
	true."

	| newCollection |
	newCollection _ self species new.
	self associationsDo: 
		[:each | 
		(aBlock value: each value) ifTrue: [newCollection add: each]].
	^newCollection! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 22:38'!
size

	^dicts
		inject: 0
		into: [ :sum :each | sum + each size ]! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 16:48'!
squeakDictionaryClass

	^Dictionary! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 17:11'!
values
	"Answer a Collection containing the receiver's values."
	| out |
	out _ WriteStream on: (Array new: self size).
	self valuesDo: [:value | out nextPut: value].
	^ out contents! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 16:57'!
valuesDo: aBlock

	dicts do: [ :each | each valuesDo: aBlock ]! !

MaDictionary subclass: #MaWeakIdentityKeyDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Squeak weak tweaks'!

!MaWeakIdentityKeyDictionary methodsFor: 'private' stamp: 'cmm 1/12/2003 21:26'!
finalizeValues

	dicts do: [ :e | e finalizeValues ]! !

!MaWeakIdentityKeyDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 22:27'!
squeakDictionaryClass

	^WeakIdentityKeyDictionary! !

MaDictionary subclass: #MaWeakValueDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Squeak weak tweaks'!

!MaWeakValueDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 22:27'!
squeakDictionaryClass

	^WeakValueDictionary! !

MaObject subclass: #MaFixedLengthRecord
	instanceVariableNames: 'byteArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-private'!

!MaFixedLengthRecord class methodsFor: 'as yet unclassified' stamp: 'cmm 5/19/2005 11:05'!
new: anInteger

	^ self new
		byteArray: (ByteArray new: anInteger) ;
		yourself! !

!MaFixedLengthRecord methodsFor: 'accessing' stamp: 'cmm 5/19/2005 10:52'!
byteArray
	^ byteArray! !

!MaFixedLengthRecord methodsFor: 'accessing' stamp: 'cmm 5/19/2005 10:52'!
byteArray: aByteArray
	byteArray _ aByteArray! !

MaFixedLengthRecord subclass: #MaHashIndexRecord
	instanceVariableNames: 'recordNumber lowKey highKey keySize valueSize numberOfSlots entrySize entryWeight canHoldDuplicateKeys'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-private'!

!MaHashIndexRecord commentStamp: 'cmm 4/28/2005 21:25' prior: 0!
keySize and valueSize are specified in bits.
entrySize is in bytes.
!

!MaHashIndexRecord class methodsFor: 'creation' stamp: 'cmm 7/18/2005 13:22'!
keySize: keyBits valueSize: valueBits numberOfSlots: entrySizeInteger

	^ self new
		keySize: keyBits
		valueSize: valueBits
		numberOfSlots: entrySizeInteger ;

		postInitialize ;
		yourself! !

!MaHashIndexRecord methodsFor: 'SUnit tests' stamp: 'cmm 12/6/2002 00:05'!
addAnomaliesTo: aDictionary usingParent: parentRecord

	| originalCount |
	originalCount _ aDictionary size.
	self
		addChildCountAnomaliesTo: aDictionary
		using: parentRecord.
	^aDictionary size > originalCount! !

!MaHashIndexRecord methodsFor: 'SUnit tests' stamp: 'cmm 12/6/2002 00:04'!
addChildCountAnomaliesTo: aDictionary using: parentRecord

	| bax parentNumberOfChildren isCorrect |
	bax _ self baxInParent: parentRecord for: self.
	parentNumberOfChildren _ (parentRecord numberOfChildrenAtByteArrayIndex: bax).
	isCorrect _ (self lowKey = self highKey or: [ self numberOfChildren = 0 ])
		ifTrue: 
			[ "it's possible for the record not to have an entry in the parent under this condition."
			(parentRecord hasEntryAtByteArrayIndex: bax)
				ifTrue: [ (parentNumberOfChildren - 1) = self numberOfChildren ]
				ifFalse: [ parentNumberOfChildren = self numberOfChildren ] ]
		ifFalse:
			[ (parentNumberOfChildren - 1) = self numberOfChildren ].
	isCorrect ifFalse:
		[ (aDictionary
			at: 'number of children discrepency'
			ifAbsentPut: [ OrderedCollection new ]) add: { parentRecord.  self } ]! !

!MaHashIndexRecord methodsFor: 'printing' stamp: ''!
allNumberOfChildren
	"This is intended to be used only for printing!!"

	^(self headerSize
		to: byteArray size - 1
		by: entrySize)
			collect: [ :index | self numberOfChildrenAtByteArrayIndex: index ]
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 9/1/2006 18:47'!
atKey: keyInteger put: valueInteger 
	"The receiver has been verified for some available slot for keyInteger."
	self 
		eligibleByteArrayIndexesFor: keyInteger
		reverseDo: 
			[ : index | 
			(self hasSlotAvailableAtByteArrayIndex: index) ifTrue: 
				[ self 
					keyAtByteArrayIndex: index
					put: keyInteger.
				self 
					valueAtByteArrayIndex: index
					put: valueInteger.
				self occupySlotAtByteArrayIndex: index.
				^ index ] ].
	MaSoftwareError signal: 'I should have been verified for an available slot.'! !

!MaHashIndexRecord methodsFor: 'SUnit tests' stamp: 'cmm 11/29/2002 23:39'!
baxInParent: parentRecord for: childRecord

	parentRecord
		byteArrayIndexesFrom: 0
		do:
			[ :bax |
			(parentRecord recordNumberAtByteArrayIndex: bax) = childRecord recordNumber
				ifTrue: [ ^bax ] ].
	self error: 'I am not child of parentRecord'! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: 'cmm 7/1/2005 17:00'!
baxMap

	| map |

	map _ OrderedCollection new.
	self
		byteArrayIndexesFrom: 0
		do: [ :each | map add: (self slotIndexForBax: each) -> each ].
	^map asArray
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 3/30/2005 10:25'!
beFrom: lowKeyInteger to: highKeyInteger

	self
		lowKey: lowKeyInteger;
		highKey: highKeyInteger.
	self setLowKey: lowKeyInteger highKey: highKeyInteger! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 7/1/2005 17:09'!
beZoomedInToward: keyInteger from: aMaHashIndexRecord
	"Set the receivers low and highIndex to the correct range that can
	accomodate the index where keyInteger belongs in aMaHashIndexRecord."

	| index |

	index _ aMaHashIndexRecord slotIndexFor: keyInteger.
	self
		beFrom: (aMaHashIndexRecord lowestKeyAtIndex: index)
		to: (aMaHashIndexRecord highestKeyAtIndex: index)
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
bufferSize

	^byteArray size
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 11/22/2002 00:12'!
byteArrayIndexContainingGlobalIndex: anInteger base: baseInteger
	"baseInteger is 0-based."

	| count |

	count _ baseInteger.
	self
		byteArrayIndexesFrom: 0
		do:
			[ :each |
			count _ count + (self numberOfChildrenAtByteArrayIndex: each).
			count >= anInteger ifTrue: [ ^each ] ].
	^nil! !

!MaHashIndexRecord methodsFor: 'actions' stamp: 'cmm 7/1/2005 16:51'!
byteArrayIndexesFrom: startingEntryIndex do: oneArgBlock

	self
		byteArrayIndexesFrom: startingEntryIndex
		to: numberOfSlots - 1
		do: oneArgBlock! !

!MaHashIndexRecord methodsFor: 'actions' stamp: ''!
byteArrayIndexesFrom: startingEntryIndex to: endingEntryIndex do: oneArgBlock

	(self byteArrayIndexFor: startingEntryIndex)
		to: (self byteArrayIndexFor: endingEntryIndex)
		by: entrySize
		do: oneArgBlock
! !

!MaHashIndexRecord methodsFor: 'private' stamp: ''!
byteArrayIndexFor: entryIndex

	^entryIndex * entrySize + self headerSize
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 11/27/2002 14:55'!
byteArrayIndexOfLowestEntry
	"Answers nil if no entries on this record."

	self
		byteArrayIndexesFrom: 0
		do:
			[ :each | (self hasEntryAtByteArrayIndex: each) ifTrue: [ ^each ] ].
	^nil! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 5/18/2005 10:11'!
cacheAttributes

	self 
		setLowKey: self lowKey
		highKey: self highKey! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
canAdd: keyInteger

	self
		eligibleByteArrayIndexesFor: keyInteger
		do: [ :index | (self hasSlotAvailableAtByteArrayIndex: index) ifTrue: [ ^true ] ].
	^false
! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
canHoldDuplicateKeys

	^canHoldDuplicateKeys! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 9/1/2006 18:11'!
controlAtByteArrayIndex: anInteger
	^ byteArray 
		maUint: self sizeOfControlFieldInBits
		at: anInteger + self controlEntryOffset! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 9/1/2006 18:12'!
controlAtByteArrayIndex: index put: anInteger
	^ byteArray 
		maUint: self sizeOfControlFieldInBits
		at: index + self controlEntryOffset
		put: anInteger! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 8/30/2006 21:08'!
controlEntryOffset
	^ 0! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 9/10/2006 20:54'!
decrementNumberOfChildrenAtByteArrayIndex: anInteger
	| currentNumberOfChildren |
	currentNumberOfChildren _ self numberOfChildrenAtByteArrayIndex: anInteger.
	currentNumberOfChildren = 0 ifTrue: [MaSoftwareError signal].
	self
		numberOfChildrenAtByteArrayIndex: anInteger
		put: currentNumberOfChildren - 1! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 11/30/2002 23:51'!
decrementNumberOfChildrenAtKey: keyInteger

	self decrementNumberOfChildrenAtByteArrayIndex: (self highestByteArrayIndexFor: keyInteger)! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: 'cmm 3/9/2004 23:53'!
dumpEntries

	self dumpEntriesTo: Transcript! !

!MaHashIndexRecord methodsFor: 'printing' stamp: 'cmm 9/10/2006 00:36'!
dumpEntriesTo: aStream 
	self 
		byteArrayIndexesFrom: 0
		do: 
			[ : eachBax | 
			(self numberOfChildrenAtByteArrayIndex: eachBax) > 0 ifTrue: 
				[ | x |
				x _ self slotIndexForBax: eachBax.
				aStream
					cr ;
					space ;
					space ;
					maPrint: 'x:' ;
					maPrint: x ;
					maPrint: ' [' ;
					maPrint: (self lowestKeyAtIndex: x) ;
					maPrint: '..' ;
					maPrint: (self highestKeyAtIndex: x) ;
					maPrint: ']' ;
					maPrint: ' bx:' ;
					maPrint: (self byteArrayIndexFor: x).
				(self hasSlotAvailableAtByteArrayIndex: eachBax) ifTrue: [ aStream nextPutAll: ' *empty slot* ' ].
				aStream
					maPrint: ' k:' ;
					maPrint: (self keyAtByteArrayIndex: eachBax) ;
					maPrint: ' v:' ;
					maPrint: (self valueAtByteArrayIndex: eachBax) ;
					maPrint: ' r:' ;
					maPrint: (self recordNumberAtByteArrayIndex: eachBax) ;
					maPrint: ' c:' ;
					maPrint: (self numberOfChildrenAtByteArrayIndex: eachBax) ;
					endEntry ] ]! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: 'cmm 4/28/2005 21:47'!
dumpStructureTo: aStream
	self
		maPrintAttributes: #(
			recordNumber lowKey highKey canHoldDuplicateKeys keys recordNumbers values
			allNumberOfChildren possibleKeys)
		on: aStream! !

!MaHashIndexRecord methodsFor: 'actions' stamp: ''!
eligibleByteArrayIndexesFor: keyInteger do: oneArgBlock

	(self lowestByteArrayIndexFor: keyInteger)
		to: (self highestByteArrayIndexFor: keyInteger)
		by: entrySize
		do: [ :index | oneArgBlock value: index ]
! !

!MaHashIndexRecord methodsFor: 'actions' stamp: 'cmm 3/10/2004 01:00'!
eligibleByteArrayIndexesFor: keyInteger reverseDo: oneArgBlock

	(self highestByteArrayIndexFor: keyInteger)
		to: (self lowestByteArrayIndexFor: keyInteger)
		by: 0-entrySize
		do: [ :index | oneArgBlock value: index ]
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 3/12/2004 15:37'!
entryWeight

	^ entryWeight! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
hasChildAtByteArrayIndex: byteArrayIndex

	^(self recordNumberAtByteArrayIndex: byteArrayIndex) > 0
! !

!MaHashIndexRecord methodsFor: 'testing' stamp: 'cmm 12/1/2002 21:44'!
hasChildAtHighestByteArrayIndexForBax: byteArrayIndex

	^(self recordNumberAtHighestByteArrayIndexForBax: byteArrayIndex) > 0! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
hasEntryAtByteArrayIndex: anInteger

	^(self hasSlotAvailableAtByteArrayIndex: anInteger) not
! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
hasEntryForKey: keyInteger
	"Answer whether keyInteger resides in an entry of the receiver."

	self
		eligibleByteArrayIndexesFor: keyInteger
		do:
			[ :index |
			((self hasEntryAtByteArrayIndex: index)
				and: [ (self keyAtByteArrayIndex: index) = keyInteger ])
					ifTrue:
						[ ^true ] ].
	^false
! !

!MaHashIndexRecord methodsFor: 'testing' stamp: 'cmm 9/1/2006 17:57'!
hasSlotAvailableAtByteArrayIndex: anInteger
	^ (self controlAtByteArrayIndex: anInteger) maIsBitClear: self slotFilledControlBit! !

!MaHashIndexRecord methodsFor: 'private' stamp: ''!
headerSize
	"Low and high keys.
	^(keySize / 8) * 2  simplified to / 4 for max performance."

	^keySize / 4
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 7/1/2005 17:00'!
highestByteArrayIndexForAnyKeyAt: byteArrayIndex
	"this is for records that canHoldDuplicateKeys when we need to find the next child record number, since that is stored at the highestByteArrayIndexFor: that key."

	^self highestByteArrayIndexFor: (self lowestKeyAtIndex: (self slotIndexForBax: byteArrayIndex))! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 7/1/2005 17:12'!
highestByteArrayIndexFor: keyInteger
	"This will vary from the lowestByteArrayIndexFor: keyInteger only when the receiver canHoldDuplicateKeys."

	^ self byteArrayIndexFor: (self highestSlotIndexFor: keyInteger)! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 3/12/2004 10:51'!
highestKeyAtIndex: anInteger

	^canHoldDuplicateKeys
		ifTrue:
			[ self lowestKeyAtIndex: anInteger ]
		ifFalse:
			[ (self lowestKeyAtIndex: anInteger + 1) - 1 ]
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 3/11/2004 21:19'!
highestPossibleKey

	^ (2 raisedTo: keySize) - 1! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 7/1/2005 17:11'!
highestSlotIndexFor: keyInteger

	^ canHoldDuplicateKeys
		ifTrue:
			[ (self slotIndexFor: keyInteger + 1) - 1 ]
		ifFalse:
			[ self slotIndexFor: keyInteger ]! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 1/27/2003 23:51'!
highKey

	^byteArray
		maUint: keySize
		at: keySize / 8
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 1/27/2003 23:57'!
highKey: anInteger

	^byteArray
		maUint: keySize
		at: (keySize / 8)
		put: anInteger
! !

!MaHashIndexRecord methodsFor: 'building' stamp: ''!
incrementNumberOfChildrenAtByteArrayIndex: anInteger

	| currentNumberOfChildren |

	currentNumberOfChildren _ self numberOfChildrenAtByteArrayIndex: anInteger.
	self
		numberOfChildrenAtByteArrayIndex: anInteger
		put: currentNumberOfChildren + 1
! !

!MaHashIndexRecord methodsFor: 'copying' stamp: 'cmm 3/24/2005 12:03'!
independentCopy

	^ self copy postIndependentCopy! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 1/27/2003 23:51'!
keyAtByteArrayIndex: anInteger

	^byteArray
		maUint: keySize
		at: anInteger + self keyEntryOffset
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 1/27/2003 23:57'!
keyAtByteArrayIndex: index put: anInteger

	^byteArray
		maUint: keySize
		at: index + self keyEntryOffset
		put: anInteger
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
keyAtHighestSlotFor: keyInteger

	^self keyAtByteArrayIndex: (self highestByteArrayIndexFor: keyInteger)
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 7/1/2005 17:13'!
keyAtSlotIndex: anInteger

	^ self keyAtByteArrayIndex: (self byteArrayIndexFor: anInteger)! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 22:44'!
keyEntryOffset

	^ self recordNumberEntryOffset + self sizeOfRecordNumber! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 3/11/2004 21:53'!
keySize

	^ keySize! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 9/1/2006 18:29'!
keySize: keyBits valueSize: valueBits numberOfSlots: numberOfSlotsInteger
	keySize _ keyBits.
	valueSize _ valueBits.
	numberOfSlots _ numberOfSlotsInteger.
	"Each entry consists of:  value, record number, key, number of child entries."
	entrySize _ self sizeOfControlFieldInBytes + (valueSize/8) + self sizeOfRecordNumber + ((keySize / 8) * 2).
	byteArray _ ByteArray new: self headerSize + (entrySize * numberOfSlotsInteger)! !

!MaHashIndexRecord methodsFor: 'printing' stamp: ''!
keys
	"This is intended to be used only for printing!!"

	^(self headerSize
		to: byteArray size - 1
		by: entrySize)
			collect: [ :index | self keyAtByteArrayIndex: index ]
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 7/1/2005 17:08'!
lowestByteArrayIndexFor: keyInteger

	^self byteArrayIndexFor: (self slotIndexFor: keyInteger)
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 7/1/2005 16:51'!
lowestKeyAtIndex: entryIndex
	"Answer the lowest-possible key that can be stored at entryIndex."

	^ lowKey + (entryIndex * entryWeight // numberOfSlots)! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 12/29/2005 22:59'!
lowestPossibleKey
	^ 0! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 1/27/2003 23:51'!
lowKey

	^byteArray
		maUint: keySize
		at: 0
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 1/27/2003 23:57'!
lowKey: anInteger

	^byteArray
		maUint: keySize
		at: 0
		put: anInteger
! !

!MaHashIndexRecord methodsFor: 'printing' stamp: 'cmm 9/10/2006 14:17'!
maPrintAbbreviatedOn: aStream

	self
		maPrintAttributes: #(recordNumber lowKey highKey entryWeight possibleKeys canHoldDuplicateKeys)
		on: aStream.
	aStream cr; maPrint: 'entries:'.
	self dumpEntriesTo: aStream! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
numberOfChildren

	| total |

	total _ 0.
	self
		byteArrayIndexesFrom: 0
		do: [ :index | total _ total + (self numberOfChildrenAtByteArrayIndex: index) ].
	^total
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 1/27/2003 23:51'!
numberOfChildrenAtByteArrayIndex: anInteger

	^byteArray
		maUint: keySize
		at: anInteger + self numberOfChildrenEntryOffset
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 1/27/2003 23:57'!
numberOfChildrenAtByteArrayIndex: index put: anInteger

	^byteArray
		maUint: keySize
		at: index + self numberOfChildrenEntryOffset
		put: anInteger! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 11/22/2002 00:13'!
numberOfChildrenBeforeByteArrayIndex: anInteger

	| total |

	total _ 0.
	self
		byteArrayIndexesFrom: 0
		do:
			[ :each |
			each = anInteger ifTrue: [ ^total ].
			total _ total + (self numberOfChildrenAtByteArrayIndex: each) ].
	self error: anInteger printString, ' is not a valid byteArrayIndex.'! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 7/1/2005 17:08'!
numberOfChildrenBefore: keyInteger

	| lastByteArrayIndex entryIndex totalInSlotsBeforeKey |

	entryIndex _ self slotIndexFor: keyInteger.
	lastByteArrayIndex _ self byteArrayIndexFor: entryIndex.
	totalInSlotsBeforeKey _ self numberOfChildrenBeforeByteArrayIndex: lastByteArrayIndex.
	^((self keyAtByteArrayIndex: lastByteArrayIndex) < keyInteger
		and: [ self hasEntryAtByteArrayIndex: lastByteArrayIndex ])
			ifTrue:
				[ (lastByteArrayIndex = (self highestByteArrayIndexFor: keyInteger))
					ifTrue: [ totalInSlotsBeforeKey + 1 ]
					ifFalse: [ totalInSlotsBeforeKey + (self numberOfChildrenAtByteArrayIndex: lastByteArrayIndex) ] ]
			ifFalse: [ totalInSlotsBeforeKey ]! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 12/30/2002 21:20'!
numberOfChildrenEntryOffset

	^self keyEntryOffset + (keySize / 8)! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 7/1/2005 16:51'!
numberOfSlots

	^ numberOfSlots! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 9/1/2006 18:45'!
occupySlotAtByteArrayIndex: anInteger 
	| controlValue |
	controlValue _ self controlAtByteArrayIndex: anInteger.
	self 
		controlAtByteArrayIndex: anInteger
		put: (controlValue maBitSet: self slotFilledControlBit)! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 3/14/2004 13:43'!
populateFrom: aMaHashIndexRecord

	byteArray
		replaceFrom: 1
		to: byteArray size
		with: aMaHashIndexRecord byteArray.
	recordNumber _ aMaHashIndexRecord recordNumber.
	self 
		setLowKey: aMaHashIndexRecord primLowKey 
		highKey: aMaHashIndexRecord primHighKey! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: 'cmm 7/1/2005 16:51'!
possibleKeys

	| answer |

	answer _ WriteStream on: (String new: numberOfSlots * 2).
	0
		to: numberOfSlots - 1
		do:
			[ : index |  | low high |
			low _ self lowestKeyAtIndex: index.
			high _ self highestKeyAtIndex: index.
			answer nextPutAll: ' | '.
			low = high
				ifTrue:
					[ answer maPrint: low ]
				ifFalse:
					[ answer
						maPrint: low;
						maPrint: '..';
						maPrint: high ] ].
	^ answer contents! !

!MaHashIndexRecord methodsFor: 'copying' stamp: 'cmm 3/24/2005 12:03'!
postIndependentCopy

	byteArray _ byteArray copy! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 12/29/2005 23:00'!
postInitialize

	self
		beFrom: self lowestPossibleKey 
		to: self highestPossibleKey! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
primHighKey

	^highKey
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
primLowKey

	^lowKey
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
recordNumber

	^recordNumber! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 1/27/2003 23:51'!
recordNumberAtByteArrayIndex: anInteger

	^byteArray 
		maUint: 32
		at: anInteger + self recordNumberEntryOffset
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 1/27/2003 23:58'!
recordNumberAtByteArrayIndex: index put: anInteger

	^byteArray
		maUint: 32
		at: index + self recordNumberEntryOffset
		put: anInteger
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 12/1/2002 20:57'!
recordNumberAtHighestByteArrayIndexForBax: anInteger

	^self recordNumberAtByteArrayIndex: (self highestByteArrayIndexForAnyKeyAt: anInteger)! !

!MaHashIndexRecord methodsFor: 'building' stamp: ''!
recordNumberAtSlotForKey: keyInteger put: recordNumberInteger

	self
		recordNumberAtByteArrayIndex: (self highestByteArrayIndexFor: keyInteger)
		put: recordNumberInteger
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
recordNumberCloserTo: keyInteger

	^self recordNumberAtByteArrayIndex: (self highestByteArrayIndexFor: keyInteger)
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 8/30/2006 21:08'!
recordNumberEntryOffset
	^ self valueEntryOffset + (valueSize/8 "bits")! !

!MaHashIndexRecord methodsFor: 'printing' stamp: ''!
recordNumbers
	"This is intended to be used only for printing!!"

	^(self headerSize
		to: byteArray size - 1
		by: entrySize)
			collect: [ :index | self recordNumberAtByteArrayIndex: index ]
! !

!MaHashIndexRecord methodsFor: 'building' stamp: ''!
recordNumber: anInteger

	recordNumber _ anInteger
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 9/10/2006 20:56'!
removeKey: keyInteger withValue: valueInteger 
	self 
		eligibleByteArrayIndexesFor: keyInteger
		do: 
			[ : each | 
			((self keyAtByteArrayIndex: each) = keyInteger 
			and: [ (self valueAtByteArrayIndex: each) = valueInteger
			and: [ self hasEntryAtByteArrayIndex: each ] ]) ifTrue: 
				[ self
					vacateSlotAtByteArrayIndex: each ;
					decrementNumberOfChildrenAtByteArrayIndex: each.
				^ each ] ].
	^ nil! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 7/1/2005 16:51'!
setLowKey: lowKeyInteger highKey: highKeyInteger

	lowKey _ self lowKey.
	highKey _ self highKey.
	canHoldDuplicateKeys _ (highKey - lowKey + 1) < numberOfSlots.
	entryWeight _ highKey - lowKey + 1! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 8/30/2006 21:17'!
sizeOfControlFieldInBits
	^ 8! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 8/30/2006 21:17'!
sizeOfControlFieldInBytes
	^ 1! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 12/30/2002 21:35'!
sizeOfRecordNumber

	^4! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 9/1/2006 17:39'!
slotFilledControlBit
	"Bit #1 of the  1-byte control field is used to mark whether a particular slot is filled or empty."
	^ 1! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 7/1/2005 16:59'!
slotIndexForBax: anInteger
	"Bax means 'byteArrayIndex'."

	^(anInteger - self headerSize) / entrySize! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 8/17/2005 19:18'!
slotIndexFor: keyInteger
	"The lowest slot index in which keyInteger can reside."

	^ (keyInteger - lowKey) * numberOfSlots // entryWeight! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: 'cmm 3/11/2004 21:41'!
structure

	^ String streamContents: [ : stream | self dumpStructureTo: stream ]! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 9/1/2006 18:45'!
vacateSlotAtByteArrayIndex: anInteger 
	| controlValue |
	controlValue _ self controlAtByteArrayIndex: anInteger.
	self 
		controlAtByteArrayIndex: anInteger
		put: (controlValue maBitClear: self slotFilledControlBit)! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 21:44'!
valueAtByteArrayIndex: anInteger

	^ byteArray 
		maUint: valueSize
		at: anInteger + self valueEntryOffset! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 21:44'!
valueAtByteArrayIndex: index put: anInteger

	^ byteArray
		maUint: valueSize
		at: index + self valueEntryOffset
		put: anInteger! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 8/30/2006 21:17'!
valueEntryOffset
	^ self controlEntryOffset + self sizeOfControlFieldInBytes! !

!MaHashIndexRecord methodsFor: 'printing' stamp: 'cmm 4/28/2005 21:52'!
values
	"This is intended to be used only for printing!!"

	^ (self headerSize
		to: byteArray size - 1
		by: entrySize)
			collect: [ : index | self valueAtByteArrayIndex: index ]! !

MaObject subclass: #MaHashIndexRecordFactory
	instanceVariableNames: 'index recordNumbersPath entryIndexPath byteArrayIndexPath recordPath preMadeNewRecord'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-private'!

!MaHashIndexRecordFactory class methodsFor: 'creation' stamp: 'cmm 12/4/2002 23:31'!
index: aMaHashIndex

	^self new
		index: aMaHashIndex;
		postInitialize;
		yourself! !

!MaHashIndexRecordFactory methodsFor: 'initializing' stamp: 'cmm 7/1/2005 17:15'!
createPreMadeNewRecord

	preMadeNewRecord :=
		MaHashIndexRecord
			keySize: index keySize
			valueSize: index valueSize
			numberOfSlots: index hashRecordSize
! !

!MaHashIndexRecordFactory methodsFor: 'accessing' stamp: ''!
entryIndexAtLevel: anInteger

	^entryIndexPath at: anInteger
! !

!MaHashIndexRecordFactory methodsFor: 'private' stamp: 'cmm 12/4/2002 23:35'!
index: aMaHashIndex

	index := aMaHashIndex! !

!MaHashIndexRecordFactory methodsFor: 'initializing' stamp: 'cmm 7/1/2005 17:15'!
initializeRecordPathRecords

	1
		to: self maxDepth
		do:
			[ :x |
			recordPath
				at: x
				put: (
					MaHashIndexRecord
						keySize: index keySize
						valueSize: index valueSize
						numberOfSlots: index hashRecordSize) ]! !

!MaHashIndexRecordFactory methodsFor: 'private' stamp: 'cmm 7/1/2005 16:54'!
maxDepth

	^(index highestPossibleKey log: index numberOfSlots) asInteger! !

!MaHashIndexRecordFactory methodsFor: 'initializing' stamp: 'cmm 12/4/2002 23:35'!
postInitialize

	recordNumbersPath := OrderedCollection new: self maxDepth.
	entryIndexPath := OrderedCollection new: self maxDepth.
	byteArrayIndexPath := OrderedCollection new: self maxDepth.
	recordPath := Array new: self maxDepth.
	self
		createPreMadeNewRecord;
		initializeRecordPathRecords! !

!MaHashIndexRecordFactory methodsFor: 'accessing' stamp: ''!
preMadeNewRecord

	^preMadeNewRecord! !

!MaHashIndexRecordFactory methodsFor: 'building' stamp: 'cmm 7/1/2005 17:00'!
put: aMaHashIndexRecord atLevel: levelInteger byteArrayIndex: indexInteger

	levelInteger > recordNumbersPath size
		ifTrue:
			[ recordNumbersPath add: aMaHashIndexRecord recordNumber.
			byteArrayIndexPath add: indexInteger.
			entryIndexPath add: (aMaHashIndexRecord slotIndexForBax: indexInteger) ]
		ifFalse:
			[ recordNumbersPath
				at: levelInteger
				put: aMaHashIndexRecord recordNumber.
			byteArrayIndexPath
				at: levelInteger
				put: indexInteger.
			entryIndexPath
				at: levelInteger
				put: (aMaHashIndexRecord slotIndexForBax: indexInteger) ].
	levelInteger > recordPath size
		ifFalse:
			[ (recordPath at: levelInteger) populateFrom: aMaHashIndexRecord ]
! !

!MaHashIndexRecordFactory methodsFor: 'accessing' stamp: 'cmm 11/29/2002 22:47'!
recordAtLevel: levelInteger

	^levelInteger > recordPath size
		ifTrue:
			[ recordPath first independentCopy ]
		ifFalse:
			[ recordPath at: levelInteger ]
! !

!MaHashIndexRecordFactory methodsFor: 'accessing' stamp: ''!
recordNumberAtLevel: anInteger

	^recordNumbersPath at: anInteger
! !

MaObject subclass: #MaIntervalCollection
	instanceVariableNames: 'intervals proximityThreshold'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Numbers collections'!

!MaIntervalCollection commentStamp: 'cmm 6/4/2005 00:48' prior: 0!
This class keeps track of a collection of intervals.

As intervals are added that intersect or are within the proximityThreshold
of existing intervals, the existing interval is enlarged to include for the new interval.

If an interval is removed in the middle of a large interval such that
a "hole" is created, a new interval will be created and the other one
will be adjusted so that they will represent only the valid ranges.

For example:

	if an instance of this class has 1..10

	and we add: 8..15, then this class will have a single interval from 1..15

	Then remove 5..11..  Assuming a proximityThreshold of 1, this class will have the following two intervals:

		1..4
		12..15

Use a proximityThreshold of 0 to track the real-numbers..!

!MaIntervalCollection methodsFor: 'add / remove' stamp: 'cmm 6/13/2005 15:43'!
addFrom: low to: high

	| intersecting lowestLow highestHigh |
	low > high ifTrue: [ self error: 'low should be lower than high' ].
	intersecting _ self intervalsIntersecting: (self expandedFrom: low to: high).
	lowestLow _ low.
	highestHigh _ high.
	intersecting do:
		[ : each |
		lowestLow _ lowestLow min: each first.
		highestHigh _ highestHigh max: each last ].
	intervals
		removeAll: intersecting ;
		add: (lowestLow to: highestHigh)! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 6/4/2005 01:04'!
expandedFrom: low to: high

	^ low-proximityThreshold to: high+proximityThreshold! !

!MaIntervalCollection methodsFor: 'testing' stamp: 'cmm 6/4/2005 01:07'!
includes: aNumber

	^ intervals anySatisfy: [ : each | each includes: aNumber ]! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 6/4/2005 01:08'!
initialize
	super initialize.
	intervals _ SortedCollection sortBlock: [ : a : b | a first < b first ].
	proximityThreshold _ 1! !

!MaIntervalCollection methodsFor: 'enumerating' stamp: 'cmm 4/12/2006 10:48'!
intervalsDo: aBlock
	intervals do: aBlock! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 6/4/2005 17:22'!
intervalsInside: anInterval

	^ intervals select: [ : each | anInterval maEnvelopes: each ]! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 6/4/2005 01:04'!
intervalsIntersecting: anInterval

	^ intervals select: [ : each | each maIntersects: anInterval ]! !

!MaIntervalCollection methodsFor: 'printing' stamp: 'cmm 6/4/2005 18:04'!
maPrintAbbreviatedOn: aStream
	super maPrintAbbreviatedOn: aStream.
	intervals do:
		[ : each |
		aStream 
			maPrint: each first ;
			maPrint: '..' ;
			maPrint: each last ;
			space ]! !

!MaIntervalCollection methodsFor: 'copying' stamp: 'cmm 10/19/2005 10:05'!
postCopy
	super postCopy.
	intervals _ intervals copy! !

!MaIntervalCollection methodsFor: 'add / remove' stamp: 'cmm 6/4/2005 18:19'!
removeFrom: low to: high
"Removing is a two-step process:
1) remove intervals completely contained by input
2) adjust intersecting intervals
	- each intersects input-low only, set its high to input-low
	- each intersects input-high only, set its low to input-high
	- each intersects input-low and input-high, set its high to inputs-low and addFrom: inputs-high to its original high"

	(self intervalsInside: (low to: high)) do: [ : each | intervals remove: each ].
	(self intervalsIntersecting: (self expandedFrom: low to: high))
		do: 
			[ : each |  | eachIntersectsLow eachIntersectsHigh |
			eachIntersectsLow _ low between: each first and: each last.
			eachIntersectsHigh _ high between: each first and: each last.
			{ eachIntersectsLow. eachIntersectsHigh } caseOf:
				{ [ { true. false } ] -> [ each setFrom: each first to: high by: each increment ].
				[ { false. true } ] -> [ each setFrom: low to: each last by: each increment ].
				[ { true. true } ] -> 
					[ | oldHigh |
					oldHigh _ each last.
					each setFrom: each first to: low by: each increment.
					self addFrom: high to: oldHigh ] } ]! !

MaObject subclass: #MaLargeArrayOfRecords
	instanceVariableNames: 'broker'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Numbers collections'!

MaLargeArrayOfRecords subclass: #MaLargeArrayOfNumbers
	instanceVariableNames: 'record'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Numbers collections'!

!MaLargeArrayOfNumbers commentStamp: 'cmm 4/28/2005 20:55' prior: 0!
MaLargeArrayOfNumbers is an auto-growing, never-shrinking array of numbers.  All numbers have the same number of bits available to represent them.  It maintains the collection in a single randomly-accessed file, and is therefore constrained by available disk space or the addressability of the file.!

!MaLargeArrayOfNumbers class methodsFor: 'utilities' stamp: 'cmm 6/14/2005 12:57'!
createNamed: filenameString bitSize: anInteger

	(FileStream isAFileNamed: filenameString)
		ifTrue: [ MaEnvironmentError signal: filenameString, ' already exists.' ].
	^ self
		create: (FileStream fileNamed: filenameString) binary
		bitSize: anInteger! !

!MaLargeArrayOfNumbers class methodsFor: 'utilities' stamp: 'cmm 6/13/2005 23:09'!
create: aReadWriteStream bitSize: anInteger

	^ (self newWithBitSize: anInteger)
		create: aReadWriteStream ;
		yourself! !

!MaLargeArrayOfNumbers class methodsFor: 'creation' stamp: 'cmm 5/23/2005 17:45'!
newWithBitSize: anInteger

	^ self new
		byteSize: anInteger / 8 ;
		yourself! !

!MaLargeArrayOfNumbers class methodsFor: 'accessing' stamp: 'cmm 3/20/2005 19:23'!
suggestedFileExtension
	^ 'idx'! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: ''!
add: anInteger

	^self
		at: self size + 1
		put: anInteger
! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 3/15/2005 21:42'!
at: index

	^ self
		at: index
		ifAbsent: [ self error: 'index out of bounds' ]! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 6/4/2005 18:38'!
at: logicalIndex ifAbsent: aBlock

	(self includesKey: logicalIndex) ifFalse: [ ^ aBlock value ].
	broker
		readInto: record
		at: logicalIndex-1.
	^ record byteArray
		maUint: broker recordSize*8
		at: 0! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 5/31/2005 15:50'!
at: logicalIndex put: anInteger

	logicalIndex > self maxSize ifTrue: [ MaUserError signal: 'Out of range.' ].
	self loadRecordWith: anInteger.
	broker
		writeFrom: record
		at: logicalIndex-1.  "-1 because broker supports 0-based indexing, I do not."
	^ anInteger! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: 'cmm 7/18/2005 11:02'!
byteSize: anInteger

	broker
		headerSize: self headerSize ;
		recordSize: anInteger.
	self initializePremadeRecord! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: ''!
headerSize
	"Header consists of:
		number of bits per Integer in the receiver."
	^1
! !

!MaLargeArrayOfNumbers methodsFor: 'testing' stamp: 'cmm 6/4/2005 18:38'!
includesKey: anInteger

	^ anInteger
		between: 1
		and: self size! !

!MaLargeArrayOfNumbers methodsFor: 'initialize' stamp: 'cmm 5/19/2005 10:09'!
initializeFromHeader: aByteArray

	broker recordSize: (
		aByteArray 
			maUint: 8
			at: 0).
	self initializePremadeRecord! !

!MaLargeArrayOfNumbers methodsFor: 'initialize' stamp: 'cmm 5/19/2005 10:08'!
initializePremadeRecord

	record _ MaFixedLengthRecord new: broker recordSize! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: 'cmm 5/19/2005 10:49'!
loadRecordWith: anInteger

	^ record byteArray
		maUint: broker recordSize * 8
		at: 0
		put: anInteger! !

!MaLargeArrayOfNumbers methodsFor: 'printing' stamp: 'cmm 2/5/2003 00:25'!
maPrintAbbreviatedOn: aStream

	1
		to: (self size min: 100)
		do:
			[ :each |
			aStream
				space;
				maPrint: (self at: each ifAbsent: [ 'error at: ', each printString ]) ].
	self size > 100 ifTrue: [ aStream nextPutAll: ' ...' ]! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: 'cmm 3/15/2005 21:40'!
maxFileSize

	^ 2000000000! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 5/19/2005 10:39'!
maxSize

	^ self maxFileSize // broker recordSize! !

!MaLargeArrayOfNumbers methodsFor: 'initialize' stamp: 'cmm 5/19/2005 10:43'!
newHeaderBytes
	"Answer a new ByteArray containing the header for the receiver.
	This is only used when creating the an instance of this class with
	the create: constructor."

	^(ByteArray new: self headerSize)
		maUint: 8
		at: 0
		put: broker recordSize;
		yourself
! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 5/31/2005 15:47'!
record
	^ record! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 5/17/2005 23:39'!
size

	^ broker size! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 4/13/2005 15:27'!
upTo: countInteger startingAt: startInteger

	| answer size targetX |
	size _ self size.
	startInteger > size ifTrue: [ ^ #() ].
	answer _ Array new: (countInteger min: size-startInteger+1).
	targetX _ 0.
	startInteger to: (startInteger+countInteger-1 min: size) do:
		[ : sourceX |
		answer
			at: (targetX _ targetX + 1)
			put: (self at: sourceX) ].
	^ answer! !

!MaLargeArrayOfRecords class methodsFor: 'create' stamp: 'cmm 6/14/2005 12:59'!
openNamed: filenameString

	(FileStream isAFileNamed: filenameString)
		ifFalse: [ MaEnvironmentError signal: filenameString, ' does not exist.' ].
	^ self open: (FileStream fileNamed: filenameString) binary! !

!MaLargeArrayOfRecords class methodsFor: 'create' stamp: 'cmm 6/14/2005 10:34'!
open: aReadWriteStream

	^ self new open: aReadWriteStream! !

!MaLargeArrayOfRecords methodsFor: 'initialize-release' stamp: 'cmm 6/14/2005 10:41'!
close

	broker close! !

!MaLargeArrayOfRecords methodsFor: 'initialize-release' stamp: 'cmm 7/11/2005 22:37'!
create: aReadWriteStream

	broker _ 
		(MaFileRecordBroker headerSize: self headerSize)
			recordSize: broker recordSize ;
			yourself.
	broker create: aReadWriteStream.
	self writeHeader! !

!MaLargeArrayOfRecords methodsFor: 'initialize-release' stamp: 'cmm 5/22/2005 21:24'!
delete

	broker delete! !

!MaLargeArrayOfRecords methodsFor: 'access' stamp: 'cmm 6/6/2005 12:31'!
filename

	^ broker filename! !

!MaLargeArrayOfRecords methodsFor: 'access' stamp: 'cmm 7/6/2005 22:52'!
filesDo: oneArgBlock

	broker filesDo: oneArgBlock! !

!MaLargeArrayOfRecords methodsFor: 'subclass responsibility' stamp: 'cmm 7/18/2005 13:27'!
headerSize

	self subclassResponsibility! !

!MaLargeArrayOfRecords methodsFor: 'initialize-release' stamp: 'cmm 5/23/2005 17:35'!
initialize

	super initialize.
	broker _ MaInImageRecordBroker new! !

!MaLargeArrayOfRecords methodsFor: 'subclass responsibility' stamp: 'cmm 5/18/2005 16:22'!
initializeFromHeader: aByteArray

	self subclassResponsibility! !

!MaLargeArrayOfRecords methodsFor: 'subclass responsibility' stamp: 'cmm 6/30/2005 23:14'!
newHeaderBytes

	self subclassResponsibility! !

!MaLargeArrayOfRecords methodsFor: 'initialize-release' stamp: 'cmm 6/14/2005 10:10'!
open: aReadWriteStream

	broker _ MaFileRecordBroker headerSize: self headerSize.
	broker open: aReadWriteStream.
	self readHeader! !

!MaLargeArrayOfRecords methodsFor: 'initialize-release' stamp: 'cmm 5/19/2005 10:13'!
readHeader

	| headerBytes |
	headerBytes _ ByteArray new: self headerSize.
	broker readHeaderInto: headerBytes.
	self initializeFromHeader: headerBytes! !

!MaLargeArrayOfRecords methodsFor: 'initialize-release' stamp: 'cmm 5/18/2005 15:53'!
writeHeader

	broker writeHeaderFrom: self newHeaderBytes! !

MaLargeArrayOfRecords subclass: #MaMagnitudeIndex
	instanceVariableNames: 'keySize valueSize nextRecordNumber highestPossibleKey'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Numbers collections'!

!MaMagnitudeIndex commentStamp: '<historical>' prior: 0!
keySize in bits.!

MaMagnitudeIndex subclass: #MaHashIndex
	instanceVariableNames: 'hashRecordSize record recordFactory preMadeKeysArrays preMadeValuesArrays guard'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Numbers collections'!

!MaHashIndex class methodsFor: 'creation' stamp: 'cmm 6/14/2005 12:57'!
createNamed: filenameString keySize: numberOfBitsInteger valueSize: valueSize recordSize: entriesInteger

	(FileStream isAFileNamed: filenameString)
		ifTrue: [ MaEnvironmentError signal: filenameString, ' already exists.' ].
	^ self
		create: (FileStream fileNamed: filenameString) binary
		keySize: numberOfBitsInteger
		valueSize: valueSize
		recordSize: entriesInteger! !

!MaHashIndex class methodsFor: 'creation' stamp: 'cmm 6/13/2005 23:07'!
create: aReadWriteStream keySize: numberOfBitsInteger valueSize: valueSize recordSize: entriesInteger

	^ (self
		newWithKeySize: numberOfBitsInteger
		valueSize: valueSize
		recordSize: entriesInteger)

		create: aReadWriteStream ;
		yourself! !

!MaHashIndex class methodsFor: 'creation' stamp: 'cmm 3/26/2006 19:43'!
newWithKeySize: numberOfBitsInteger valueSize: valueSize recordSize: entriesInteger
	"Use this constructor only when you want to create in-memory cached-hash-index."
	(self validRecordSizes includes: entriesInteger) 
		ifFalse: [ MaUserError signal: 'recordSize must be one of ', MaHashIndex validRecordSizes printString ].
	^ self new
		keySize: numberOfBitsInteger
		valueSize: valueSize
		hashRecordSize: entriesInteger ;

		writeNewRoot ;
		yourself! !

!MaHashIndex class methodsFor: 'creation' stamp: ''!
suggestedFileExtension

	^'hdx'
! !

!MaHashIndex class methodsFor: 'accessing' stamp: 'cmm 8/17/2005 21:21'!
validRecordSizes
	"Record sizes must be a power of 2 so that slots in every MaHashIndexRecord will be uniformely sized."

	^ #(2 4 8 16 32 64 128)! !

!MaHashIndex methodsFor: 'add / remove' stamp: 'cmm 9/1/2006 18:45'!
add: valueInteger at: keyInteger 
	"Find the record which either has space for keyInteger or is as far zoomed-in as possible (with either space for keyInteger or not).  If a record already contains another key greater than keyInteger, that entry must be 'pushed' down so we can maintain key order.  Additionally, we must increment the numberOfChildren for each record to the path of where we finally add."
	guard critical: 
		[ | currentKeyToWrite currentValueToWrite hasSpace reachedBottom |
		self 
			validateCanAdd: valueInteger
			at: keyInteger.
		currentKeyToWrite _ keyInteger.
		currentValueToWrite _ valueInteger.
		self selectRoot.
		hasSpace _ false.
		reachedBottom _ false.
		[ hasSpace | reachedBottom ] whileFalse: 
			[ | highestBax keyToMove valueToMove childRecordNumber |
			highestBax _ record highestByteArrayIndexFor: currentKeyToWrite.
			((hasSpace _ record canAdd: currentKeyToWrite) or: 
				[ (keyToMove _ record keyAtHighestSlotFor: currentKeyToWrite) > currentKeyToWrite ]) 
				ifTrue: 
					[ hasSpace ifFalse: 
						[ valueToMove _ record valueAtByteArrayIndex: highestBax.
						record vacateSlotAtByteArrayIndex: highestBax ].
					"Incrementing here is 'corrected', below, if necessary.."
					record incrementNumberOfChildrenAtByteArrayIndex: (record 
							atKey: currentKeyToWrite
							put: currentValueToWrite).
					hasSpace ifFalse: 
						[ "It should be safe to assume, here, that highestBax is where currentKeyToWrite was just placed."
						(record canAdd: keyToMove) 
							ifTrue: 
								[ record
									incrementNumberOfChildrenAtByteArrayIndex: (record 
											atKey: keyToMove
											put: valueToMove) ;
									decrementNumberOfChildrenAtByteArrayIndex: highestBax.
								hasSpace _ true ]
							ifFalse: 
								[ "If the displaced entry will not be a child of the displacer, we need to correct the increment, (see) above, by decrement the numberOfChildren and incrementing where it *will* be placed."
								| newBax |
								newBax _ record highestByteArrayIndexFor: keyToMove.
								highestBax = newBax ifFalse: 
									[ record
										decrementNumberOfChildrenAtByteArrayIndex: highestBax ;
										incrementNumberOfChildrenAtByteArrayIndex: newBax ] ] ] ]
				ifFalse: [ record incrementNumberOfChildrenAtByteArrayIndex: highestBax ].
			self writeRecord: record.
			valueToMove notNil ifTrue: 
				[ "we will need to push this key/value pair down"
				currentKeyToWrite _ keyToMove.
				currentValueToWrite _ valueToMove ].
			childRecordNumber _ record recordNumberCloserTo: currentKeyToWrite.
			reachedBottom _ childRecordNumber = 0.
			reachedBottom ifFalse: [ self selectRecord: childRecordNumber ] ].
		(reachedBottom and: [ hasSpace not ]) ifTrue: 
			[ self selectNewRecordFor: currentKeyToWrite.
			record incrementNumberOfChildrenAtByteArrayIndex: (record 
					atKey: currentKeyToWrite
					put: currentValueToWrite).
			self writeRecord ] ]! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 6/12/2006 22:48'!
allRecordsDo: twoArgBlock 
	guard critical: 
		[ self selectRoot.
		self 
			recordsFrom: record independentCopy
			do: twoArgBlock ]! !

!MaHashIndex methodsFor: 'add / remove' stamp: 'cmm 3/26/2006 19:22'!
at: key put: value
	"For collection api compatibility."
	self
		add: value
		at: key.
	^ value! !

!MaHashIndex methodsFor: 'initializing' stamp: 'cmm 7/18/2005 13:22'!
createRecord

	record _
		(MaHashIndexRecord
			keySize: keySize
			valueSize: valueSize
			numberOfSlots: hashRecordSize)

				beFrom: 0
				to: (2 raisedTo: keySize) - 1;

				recordNumber: 0 ;
				yourself.
	broker recordSize: record bufferSize! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 3/22/2005 18:05'!
do: aBlock for: aMagmaFileTraverser using: aMaReadStrategy

	self
		entriesFromIndex: 1
		do:
			[ :eachRecord :recordPosition :levels |
			aMagmaFileTraverser
				graphFrom: (eachRecord keyAtByteArrayIndex: recordPosition)
				do: aBlock
				using: aMaReadStrategy ]
		until:
			[ :eachRecord :recordPosition :levels | false ]! !

!MaHashIndex methodsFor: 'enumerating' stamp: ''!
entriesFromIndex: globalIndex do: threeArgBlock until: conditionBlock

	^self
		slotsFromIndex: globalIndex
		do:
			[ :currentRecord :byteArrayIndex :level |
			(currentRecord hasEntryAtByteArrayIndex: byteArrayIndex)
				ifTrue:
					[ threeArgBlock
						value: currentRecord
						value: byteArrayIndex
						value: level ] ]
		until: conditionBlock
! !

!MaHashIndex methodsFor: 'enumerating' stamp: ''!
entriesFrom: keyInteger do: threeArgBlock until: conditionBlock
	"The first two arguments of threeArgBlock and conditionBlock are the record and the
	internal index position needed to access the of each entry in those records via
	keyAtByteArrayIndex:, oidAtByteArrayIndex:, and recordNumberAtByteArrayIndex:.
	The third argument is the levels from the root record, where the root is level 1.
	This is used to reference into the receivers preMadeRecords, a performance optimization."

	^self
		slotsFrom: keyInteger
		do:
			[ :currentRecord :byteArrayIndex :level |
			(currentRecord hasEntryAtByteArrayIndex: byteArrayIndex)
				ifTrue:
					[ threeArgBlock
						value: currentRecord
						value: byteArrayIndex
						value: level ] ]
		until: conditionBlock
! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 9/9/2006 23:42'!
findAnomalies
	| answer |
	answer _ Dictionary new.
	self allRecordsDo:
		[ :eachParentRec :eachChildRec |
		(eachChildRec 
			addAnomaliesTo: answer 
			usingParent: eachParentRec) ifTrue:
				[ MaNotification new
					tag: answer;
					signal ] ].
	^answer! !

!MaHashIndex methodsFor: 'private' stamp: ''!
getNextRecordNumber

	^nextRecordNumber _ nextRecordNumber + 1
! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 7/2/2005 17:06'!
hasAnomalies

	^ self findAnomalies notEmpty! !

!MaHashIndex methodsFor: 'accessing' stamp: ''!
hashRecordSize

	^hashRecordSize! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 4/29/2005 12:13'!
headerSize
	"Header consists of:
		keySize:  8 bits
		valueSize: 8 bits
		recordSize: 8 bits"

	^ 3
! !

!MaHashIndex methodsFor: 'testing' stamp: 'cmm 6/12/2006 22:50'!
includesKey: anInteger 
	^ guard critical: 
		[ self selectFirstRecordWithKeyEqualOrGreaterThan: anInteger.
		record hasEntryForKey: anInteger ]! !

!MaHashIndex methodsFor: 'testing' stamp: 'cmm 4/28/2005 21:55'!
includesValue: valueInteger at: keyInteger

	| hasEntry key |
	self
		entriesFrom: keyInteger
		do:
			[ : eachRecord : eachByteArrayIndex : level |
			(hasEntry and:
				[ key = keyInteger
					and: [ (eachRecord valueAtByteArrayIndex: eachByteArrayIndex) = valueInteger ] ])
						ifTrue:
							[ ^ true ] ]
		until:
			[ : eachRecord : eachByteArrayIndex : level |
			"Until we're on an entry that could not possibly hold keyInteger."
			((hasEntry _ eachRecord hasEntryAtByteArrayIndex: eachByteArrayIndex)
				and: [ (key _ eachRecord keyAtByteArrayIndex: eachByteArrayIndex) > keyInteger ])
					or: [ eachByteArrayIndex > (eachRecord highestByteArrayIndexFor: keyInteger) ] ].
	^ false! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 6/12/2006 17:27'!
indexOfHighest: keyInteger 
	"If keyInteger is present, answer the index of its last occurrence.  Otherwise, answer the index of the next lower key.  If keyInteger is lower than all keys in the receiver, answer 0."
	| lowestIndexOfNextHigherKey |
	^ guard critical: 
		[ keyInteger >= self highestPossibleKey ifTrue: [ ^ self size ].
		lowestIndexOfNextHigherKey _ self 
			indexOfLowest: keyInteger + 1
			exactKey: false.
		lowestIndexOfNextHigherKey = 1 ifTrue: [ ^ 0 ].	"lower than all"
		lowestIndexOfNextHigherKey = 0 
			ifTrue: [ self size ]
			ifFalse: [ lowestIndexOfNextHigherKey - 1 max: 1 ] ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 6/12/2006 17:26'!
indexOfLowest: keyInteger exactKey: aBoolean 
	"If aBoolean is true, answer 0 if we don't find an exact key.  If false answer the index of the first equal or higher key."
	| total index |
	^ guard critical: 
		[ total _ 0.
		self 
			navigateToward: keyInteger
			until: 
				[ : currentRecord | 
				| foundRecord |
				foundRecord _ false.
				currentRecord 
					eligibleByteArrayIndexesFor: keyInteger
					do: 
						[ : each | 
						((currentRecord hasEntryAtByteArrayIndex: each) and: [ (currentRecord keyAtByteArrayIndex: each) >= keyInteger ]) ifTrue: [ foundRecord _ true ] ].
				total _ total + (record numberOfChildrenBefore: keyInteger).
				foundRecord ]
			do: 
				[ : currentRecord | "nada" ].
		"We now have the correct potential record selected and a correct total number of entries that occur before the entry that would contain keyInteger.  Now determine whether we should answer it or nil"
		index _ total + 1.
		aBoolean 
			ifTrue: 
				[ (record hasEntryForKey: keyInteger) 
					ifTrue: [ ^ index ]
					ifFalse: [ 0 ] ]
			ifFalse: 
				[ index > self size 
					ifTrue: [ 0 ]
					ifFalse: [ index ] ] ]! !

!MaHashIndex methodsFor: 'accessing' stamp: ''!
indexOf: keyInteger

	^self
		indexOfLowest: keyInteger
		exactKey: true
! !

!MaHashIndex methodsFor: 'initializing' stamp: 'cmm 6/12/2006 17:22'!
initialize

	super initialize.
	self initializePreMadeAnswerArrays.
	guard _ Monitor new! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 6/14/2005 13:22'!
initializeFromHeader: aByteArray

	self
		keySize: (aByteArray maUint: 8 at: 0)+1
		valueSize: (aByteArray maUint: 8 at: 1)+1
		hashRecordSize: (aByteArray maUint: 8 at: 2).
	broker recordSize: record bufferSize! !

!MaHashIndex methodsFor: 'initializing' stamp: 'cmm 4/17/2005 22:01'!
initializePreMadeAnswerArrays

	preMadeKeysArrays _ (1 to: 250) collect: [ :eachSize | Array new: eachSize ].
	preMadeValuesArrays _ (1 to: 250) collect: [ :eachSize | Array new: eachSize ]! !

!MaHashIndex methodsFor: 'initializing' stamp: ''!
initializeRecordFactory

	recordFactory _ MaHashIndexRecordFactory index: self
! !

!MaHashIndex methodsFor: 'initializing' stamp: ''!
initializeRecords

	self
		createRecord;
		initializeRecordFactory
! !

!MaHashIndex methodsFor: 'initializing' stamp: 'cmm 5/23/2005 17:54'!
keySize: keySizeInteger valueSize: valueSizeInteger hashRecordSize: recordSizeInteger

	keySize _ keySizeInteger.
	valueSize _ valueSizeInteger.
	hashRecordSize _ recordSizeInteger.
	self initializeRecords! !

!MaHashIndex methodsFor: 'enumerating' stamp: 'cmm 6/4/2006 12:53'!
keysAndValuesFrom: lowKey startingAt: anInteger do: twoArgBlock until: conditionBlock 
	"Value twoArgBlock with each key and value pair, until conditionBlock is false for same key-value pair, starting at anInteger after the lowest index of lowKey."
	| lowest currentIndex |
	currentIndex _ 0.
	lowest _ self 
		indexOfLowest: lowKey
		exactKey: false.
	lowest > 0 ifTrue: 
		[ self 
			entriesFromIndex: lowest + anInteger - 1
			do: 
				[ : currentRecord : byteArrayIndex : level | 
				currentIndex _ currentIndex + 1.
				twoArgBlock 
					value: (currentRecord keyAtByteArrayIndex: byteArrayIndex)
					value: (currentRecord valueAtByteArrayIndex: byteArrayIndex) ]
			until: 
				[ : currentRecord : byteArrayIndex : level | 
				conditionBlock 
					value: (currentRecord keyAtByteArrayIndex: byteArrayIndex)
					value: (currentRecord valueAtByteArrayIndex: byteArrayIndex) ] ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 4/17/2005 22:02'!
keysArrayOfSize: anInteger

	^anInteger > preMadeKeysArrays size
		ifTrue:
			[ Array new: anInteger ]
		ifFalse:
			[ preMadeKeysArrays at: anInteger ]
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/27/2002 13:22'!
navigateToward: keyInteger until: untilBlock do: doBlock
	"Select records between the root and the one that can contain keyInteger
	and value one-argument doBlock on each of those records until the
	one-argument untilBlock values true or we cannot navigate any
	closer to keyInteger."

	| level |

	level _ 1.
	self selectRoot.
	[ | goalRecordReached bax |
	bax _ record highestByteArrayIndexFor: keyInteger.
	recordFactory
		put: record
		atLevel: level
		byteArrayIndex: bax.
	goalRecordReached _ untilBlock value: record.
	goalRecordReached not and: 	[ record hasChildAtByteArrayIndex: bax ] ]
		whileTrue:
			[ doBlock value: record.
			self selectRecord: (record recordNumberCloserTo: keyInteger).
			level _ level + 1 ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 6/14/2005 13:22'!
newHeaderBytes

	^ (ByteArray new: self headerSize)
		maUint: 8
		at: 0
		put: keySize-1 ;   "-1 because we want to allow up to 256 bits"

		maUint: 8
		at: 1
		put: valueSize-1 ;

		maUint: 8
		at: 2
		put: hashRecordSize ;

		yourself! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 6/12/2006 17:28'!
numberOfEntriesFrom: lowKey to: highKey 
	"Inclusive of lowKey and highKey, if those keys exist.
	If lowKey does not exist, we count from the next higher key.
	If highKey doesn't exist, we count up to the next highestKey."
	^ guard critical: 
		[ | lowest |
		(lowest _ self 
			indexOfLowest: lowKey
			exactKey: false) = 0 ifTrue: [ ^ 0 ].
		(self indexOfHighest: highKey) - lowest + 1 ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 7/1/2005 16:54'!
numberOfSlots

	^ record numberOfSlots! !

!MaHashIndex methodsFor: 'accessing' stamp: ''!
occurrencesOf: keyInteger

	^self
		numberOfEntriesFrom: keyInteger
		to: keyInteger
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 5/22/2005 22:54'!
readRecord: aMaHashIndexRecord at: recordIndex

	broker
		readInto: aMaHashIndexRecord
		at: recordIndex.
	^ aMaHashIndexRecord 
		cacheAttributes ;
		recordNumber: recordIndex ;
		yourself! !

!MaHashIndex methodsFor: 'private' stamp: ''!
recordSize

	^record bufferSize
! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 6/12/2006 22:49'!
recordsDo: twoArgBlock level: level byteArrayIndex: anInteger 
	guard critical: 
		[ | parent |
		recordFactory 
			put: record
			atLevel: level
			byteArrayIndex: anInteger.
		parent _ recordFactory recordAtLevel: level.
		twoArgBlock 
			value: parent
			value: level.
		parent 
			byteArrayIndexesFrom: 0
			do: 
				[ : eachBax | 
				(parent hasChildAtByteArrayIndex: eachBax) ifTrue: 
					[ self selectRecord: (parent recordNumberAtByteArrayIndex: eachBax).
					self 
						recordsDo: twoArgBlock
						level: level + 1
						byteArrayIndex: anInteger ] ] ]! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 6/12/2006 22:49'!
recordsFrom: aMaHashIndexRecord do: twoArgBlock 
	guard critical: 
		[ | childRec |
		aMaHashIndexRecord 
			byteArrayIndexesFrom: 0
			do: 
				[ : eachBax | 
				(aMaHashIndexRecord hasChildAtByteArrayIndex: eachBax) ifTrue: 
					[ childRec _ aMaHashIndexRecord independentCopy.
					self 
						readRecord: childRec
						at: (aMaHashIndexRecord recordNumberAtByteArrayIndex: eachBax).
					twoArgBlock 
						value: aMaHashIndexRecord
						value: childRec.
					self 
						recordsFrom: childRec
						do: twoArgBlock ] ] ]! !

!MaHashIndex methodsFor: 'add / remove' stamp: 'cmm 9/1/2006 18:45'!
remove: valueInteger at: keyInteger 
	"Answer a boolean indicating whether anything was removed."
	guard critical: 
		[ | currentBaxToFill sourceBax parentRecord childRecord baxFilled |
		(self 
			includesValue: valueInteger
			at: keyInteger) ifFalse: [ ^ false ].
		self 
			navigateToward: keyInteger
			until: 
				[ : eachRecord | 
				(currentBaxToFill _ eachRecord 
					removeKey: keyInteger
					withValue: valueInteger) notNil ]
			do: 
				[ : eachRecord | 
				"No need for beforeImageSignal here because we already signaled it in the until block."
				eachRecord decrementNumberOfChildrenAtKey: keyInteger.
				self writeRecord ].
		self writeRecord.

		"get out if we can.."
		currentBaxToFill ifNil: [ ^ true ].
		(record hasChildAtByteArrayIndex: currentBaxToFill) ifFalse: [ ^ true ].

		"Now pull up child record entries into the newly vacated slot because add looks for the first available slot that is eligible for the key it is adding, which *could* be greater than the key of the next child (meaning we'd then be out of order)."
		sourceBax _ 0.	"non-nil"
		parentRecord _ record independentCopy.
		childRecord _ record independentCopy.
		
		[ sourceBax notNil and: [ parentRecord hasChildAtHighestByteArrayIndexForBax: currentBaxToFill ] ] whileTrue: 
			[ | highestByteArrayIndexForBaxToFill |
			highestByteArrayIndexForBaxToFill _ parentRecord highestByteArrayIndexForAnyKeyAt: currentBaxToFill.
			self 
				readRecord: childRecord
				at: (parentRecord recordNumberAtByteArrayIndex: highestByteArrayIndexForBaxToFill).
			sourceBax _ childRecord byteArrayIndexOfLowestEntry.
			sourceBax ifNotNil: 
				[ baxFilled _ parentRecord 
					atKey: (childRecord keyAtByteArrayIndex: sourceBax)
					put: (childRecord valueAtByteArrayIndex: sourceBax).
				baxFilled ~= highestByteArrayIndexForBaxToFill ifTrue: 
					[ parentRecord incrementNumberOfChildrenAtByteArrayIndex: baxFilled.
					parentRecord decrementNumberOfChildrenAtByteArrayIndex: highestByteArrayIndexForBaxToFill ].
				childRecord
					decrementNumberOfChildrenAtByteArrayIndex: sourceBax ;
					vacateSlotAtByteArrayIndex: sourceBax.
				self
					writeRecord: parentRecord ;
					writeRecord: childRecord.
				parentRecord populateFrom: childRecord.
				currentBaxToFill _ sourceBax ] ].
		^ true ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 7/1/2005 17:10'!
selectFirstRecordWithKeyEqualOrGreaterThan: keyInteger
	"If keyInteger is larger than every key in the index, answer nil, otherwise answer the level of the record with the answer."

	| level searchedLevel |
	level _ 1.
	self
		navigateToward: keyInteger
		until:
			[ : currentRecord |
			currentRecord
				eligibleByteArrayIndexesFor: keyInteger
				do:
					[ : each |
					((currentRecord hasEntryAtByteArrayIndex: each)
						and: [ (currentRecord keyAtByteArrayIndex: each) >= keyInteger ])
							ifTrue:
								[ ^ level ] ].
			false ]
		do: [ :currentRecord | level _ level + 1 ].
	"Though we have arrived at the record which would contain keyInteger,
	in case it does not have entries whose keys are >= keyInteger,
	look for the next higher key."
	self
		traverseSlotsFromRecord: record
		index: (record slotIndexFor: keyInteger)
		do: [ : currentRecord : byteArrayIndex : currentLevel | "nothing" ]
		until:
			[ : currentRecord : byteArrayIndex : currentLevel |
			searchedLevel _ currentLevel.
			currentRecord hasEntryAtByteArrayIndex: byteArrayIndex ]
		level: level.
	^ searchedLevel! !

!MaHashIndex methodsFor: 'private' stamp: ''!
selectNewRecordFor: keyInteger
	"We need to make sure we reference the new record from the
	current record."

	self setUpNewRecordFor: keyInteger.
	"Reference new record and write before selecting the new record."
	record
		recordNumberAtSlotForKey: keyInteger
		put: recordFactory preMadeNewRecord recordNumber.
	self
		writeRecord; "replace record that now references newRecord.."
		selectPreMadeNewRecord
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
selectPreMadeNewRecord

	record populateFrom: recordFactory preMadeNewRecord
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/30/2002 13:03'!
selectRecordAt: globalIndex
	"Answer the level and the byteArrayIndex of the selected record, or nil if out of bounds.
	Since the caller of this method is going to call traverseSlotsFromRecord:index:do:until:level:,
	we must be sure to populate the receivers path."

	| level byteArrayIndex currentGlobalPosition outOfBounds |

	self selectRoot.
	level _ 1.
	currentGlobalPosition _ 0.  "0-based"
	outOfBounds _ false.
	[
		byteArrayIndex _
			record
				byteArrayIndexContainingGlobalIndex: globalIndex
				base: currentGlobalPosition.
		byteArrayIndex isNil
			ifTrue:
				[ outOfBounds _ true.
				false ]
			ifFalse:
				[ currentGlobalPosition _
					currentGlobalPosition +
						(record numberOfChildrenBeforeByteArrayIndex: byteArrayIndex).
				(record hasEntryAtByteArrayIndex: byteArrayIndex)
					ifTrue: [ currentGlobalPosition _ currentGlobalPosition + 1 ].
				(currentGlobalPosition < globalIndex) and:
					[ "this should never be false if above is true"
					record hasChildAtByteArrayIndex: byteArrayIndex ] ] ]
						whileTrue:
							[ recordFactory
								put: record
								atLevel: level
								byteArrayIndex: byteArrayIndex.
							self selectRecord:
								(record recordNumberAtByteArrayIndex: byteArrayIndex).
							level _ level + 1 ].
	^outOfBounds
		ifTrue:
			[ nil ]
		ifFalse:
			[ Array
				with: byteArrayIndex
				with: level ]! !

!MaHashIndex methodsFor: 'private' stamp: ''!
selectRecord: anInteger

	self
		readRecord: record
		at: anInteger
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
selectRoot

	self selectRecord: 0
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
setUpNewRecordFor: keyInteger

	recordFactory preMadeNewRecord
		beZoomedInToward: keyInteger
			from: record;
		recordNumber: self getNextRecordNumber
! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 6/12/2006 17:26'!
size
	^ guard critical: 
		[ self selectRoot.
		record numberOfChildren ]! !

!MaHashIndex methodsFor: 'enumerating' stamp: 'cmm 6/12/2006 22:46'!
slotsFromIndex: globalIndex do: threeArgBlock until: conditionBlock 
	"The first two arguments of threeArgBlock and conditionBlock are the record and the internal index position needed to access the of each entry in those records via keyAtByteArrayIndex:, oidAtByteArrayIndex:, and recordNumberAtByteArrayIndex:.  The third argument is the levels from the root record, where the root is level 1.  This is used to reference into the receivers preMadeRecords, a performance optimization."
	guard critical: 
		[ | byteArrayIndexAndLevelArray |
		(byteArrayIndexAndLevelArray _ self selectRecordAt: globalIndex) isNil ifTrue: [ ^ self ].
		self 
			traverseSlotsFromRecord: record
			index: (record slotIndexForBax: byteArrayIndexAndLevelArray first)
			do: threeArgBlock
			until: conditionBlock
			level: byteArrayIndexAndLevelArray last ]! !

!MaHashIndex methodsFor: 'enumerating' stamp: 'cmm 6/12/2006 22:46'!
slotsFrom: keyInteger do: threeArgBlock until: conditionBlock 
	"The first two arguments of threeArgBlock and conditionBlock are the record and the internal index position needed to access the of each entry in those records via keyAtByteArrayIndex:, oidAtByteArrayIndex:, and recordNumberAtByteArrayIndex:.  The third argument is the levels from the root record, where the root is level 1.  This is used to reference into the receivers preMadeRecords, a performance optimization."
	guard critical: 
		[ | level startingEntryIndex |
		(level _ self selectFirstRecordWithKeyEqualOrGreaterThan: keyInteger) isNil ifTrue: [ ^ self ].
		"For inverted records, it is possible the entryIndex for keyInteger will be consumed by the next lower key.  Unfortunately, we have to check."
		startingEntryIndex _ record slotIndexFor: keyInteger.
		(record keyAtSlotIndex: startingEntryIndex) < keyInteger ifTrue: [ startingEntryIndex _ startingEntryIndex + 1 ].
		self 
			traverseSlotsFromRecord: record
			index: startingEntryIndex
			do: threeArgBlock
			until: conditionBlock
			level: level ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/29/2002 22:46'!
traverseSlotsFromParentRecord: aMaHashIndexRecord index: entryIndex do: threeArgBlock until: conditionBlock level: levelInteger
	"levelInteger is 1 for the root record."

	| conditionMet |

	conditionMet _ false.
	aMaHashIndexRecord
		byteArrayIndexesFrom: entryIndex
		do:
			[ :index | | nextToRead |
			conditionMet
				ifFalse:
					[ nextToRead _ aMaHashIndexRecord recordNumberAtByteArrayIndex: index.
					(conditionBlock
						value: aMaHashIndexRecord
						value: index
						value: levelInteger)
							ifTrue:
								[ conditionMet _ true ]
							ifFalse:
								[ threeArgBlock
									value: aMaHashIndexRecord
									value: index
									value: levelInteger ] ].
			conditionMet
				ifFalse:
					[ nextToRead > 0
						ifTrue:
							[ | currentRecord |
							currentRecord _ recordFactory recordAtLevel: levelInteger + 1.
							self
								readRecord: currentRecord
								at: nextToRead.
							conditionMet _
								self
									traverseSlotsFromParentRecord: currentRecord
									index: 0
									do: threeArgBlock
									until: conditionBlock
									level: levelInteger + 1 ] ] ].
	^conditionMet
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/29/2002 22:46'!
traverseSlotsFromRecord: aMaHashIndexRecord index: entryIndex do: threeArgBlock until: conditionBlock level: levelInteger
	"levelInteger is 1 for the root record."

	| done |

	done _
		self
			traverseSlotsFromParentRecord: aMaHashIndexRecord
			index: entryIndex
			do: threeArgBlock
			until: conditionBlock
			level: levelInteger.
	(done not and: [ levelInteger > 1 ])
		ifTrue:
			[ | nextStart |
			nextStart _ recordFactory recordAtLevel: levelInteger - 1.
			self
				readRecord: nextStart
				at: (recordFactory recordNumberAtLevel: levelInteger - 1).
			self
				traverseSlotsFromRecord: nextStart
				index: (recordFactory entryIndexAtLevel: levelInteger - 1) + 1
				do: threeArgBlock
				until: conditionBlock
				level: levelInteger - 1 ]
! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 4/17/2005 22:13'!
upTo: countInteger at: keyInteger

	^ self
		upTo: countInteger
		keysAndValuesFrom: keyInteger
		to: keyInteger
		startingAt: 1! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 6/12/2006 17:30'!
upTo: countInteger keysAndValuesFromIndex: lowIndex 
	^ guard critical: 
		[ | answerIndex keys values |
		keys _ self keysArrayOfSize: countInteger.
		values _ self valuesArrayOfSize: countInteger.
		answerIndex _ 0.
		self 
			entriesFromIndex: lowIndex
			do: 
				[ : currentRecord : byteArrayIndex : level | 
				keys 
					at: (answerIndex _ answerIndex + 1)
					put: (currentRecord keyAtByteArrayIndex: byteArrayIndex).
				values 
					at: answerIndex
					put: (currentRecord valueAtByteArrayIndex: byteArrayIndex) ]
			until: [ : currentRecord : index : level | answerIndex = countInteger ].
		answerIndex = 0 
			ifTrue: 
				[ {  #(). #()  } ]
			ifFalse: 
				[ answerIndex < countInteger 
					ifTrue: 
						[ { 
							(keys 
								copyFrom: 1
								to: answerIndex).
							(values 
								copyFrom: 1
								to: answerIndex)
						 } ]
					ifFalse: 
						[ {  keys. values  } ] ] ]! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 6/12/2006 17:29'!
upTo: countInteger keysAndValuesFrom: lowKey to: highKey startingAt: index 
	"Answer the specified number of values starting at logical-position index *within* the lowKey to highKey range."
	^ guard critical: 
		[ | answerIndex keys values |
		answerIndex _ 0.
		keys _ self keysArrayOfSize: countInteger.
		values _ self valuesArrayOfSize: countInteger.
		self 
			upTo: countInteger
			keysAndValuesFrom: lowKey
			to: highKey
			startingAt: index
			do: 
				[ : eachKey : eachValue | 
				keys 
					at: (answerIndex _ answerIndex + 1)
					put: eachKey.
				values 
					at: answerIndex
					put: eachValue ].
		answerIndex = 0 
			ifTrue: 
				[ {  #(). #()  } ]
			ifFalse: 
				[ answerIndex < countInteger 
					ifTrue: 
						[ { 
							(keys 
								copyFrom: 1
								to: answerIndex).
							(values 
								copyFrom: 1
								to: answerIndex)
						 } ]
					ifFalse: 
						[ {  keys. values  } ] ] ]! !

!MaHashIndex methodsFor: 'enumerating' stamp: 'cmm 4/12/2006 21:53'!
upTo: countInteger keysAndValuesFrom: lowKey to: highKey startingAt: index do: twoArgBlock
	"Answer the specified number of values starting at logical-position index within the lowKey to highKey range, value twoArgBlock with each key-value pair in that range."
	| lowest currentIndex |
	currentIndex _ 0.
	lowest _ self indexOfLowest: lowKey exactKey: false.
	lowest > 0
		ifTrue:
			[ self
				entriesFromIndex: lowest + index - 1
				do:
					[ : currentRecord : byteArrayIndex : level |
					currentIndex _ currentIndex + 1.
					twoArgBlock
						value: (currentRecord keyAtByteArrayIndex: byteArrayIndex)
						value: (currentRecord valueAtByteArrayIndex: byteArrayIndex) ]
				until:
					[  : currentRecord : byteArrayIndex : level |
					currentIndex = countInteger
					or: [ (currentRecord keyAtByteArrayIndex: byteArrayIndex) > highKey ] ] ]! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 6/12/2006 22:49'!
upTo: countInteger keysFrom: startKey 
	guard critical: 
		[ | result answerIndex |
		result _ self valuesArrayOfSize: countInteger.
		answerIndex _ 1.
		self 
			entriesFrom: startKey
			do: 
				[ : currentRecord : index : level | 
				(currentRecord hasEntryAtByteArrayIndex: index) ifTrue: 
					[ result 
						at: answerIndex
						put: (currentRecord keyAtByteArrayIndex: index).
					answerIndex _ answerIndex + 1 ] ]
			until: [ : currentRecord : index : level | answerIndex > countInteger ].
		"Need to check in case of a small index (say only 9 entries?)."
		^ answerIndex < countInteger 
			ifTrue: 
				[ result 
					copyFrom: 1
					to: answerIndex - 1 ]
			ifFalse: [ result ] ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 3/20/2005 17:49'!
validateCanAdd: oidInteger at: keyInteger

	keyInteger > self highestPossibleKey ifTrue:
		[ MaHashIndexUserError signal: keyInteger printString, ' is larger than the highestPossibleKey.' ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 4/17/2005 22:01'!
valuesArrayOfSize: anInteger

	^anInteger > preMadeValuesArrays size
		ifTrue:
			[ Array new: anInteger ]
		ifFalse:
			[ preMadeValuesArrays at: anInteger ]
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
writeNewRoot

	self writeRecord
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
writeRecord

	self writeRecord: record
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 5/22/2005 22:54'!
writeRecord: aMaHashIndexRecord

	broker 
		writeFrom: aMaHashIndexRecord
		at: aMaHashIndexRecord recordNumber! !

!MaMagnitudeIndex methodsFor: 'building' stamp: ''!
add: oidInteger at: keyInteger

	self subclassResponsibility
! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: ''!
headerSize

	self subclassResponsibility
! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: 'cmm 3/20/2005 17:50'!
highestPossibleKey

	^ highestPossibleKey ifNil:
		[ highestPossibleKey _ (2 raisedTo: keySize) - 1 ]
! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: 'cmm 5/23/2005 23:13'!
initialize
	super initialize.
	nextRecordNumber _ 0! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: ''!
keySize

	^keySize! !

!MaMagnitudeIndex methodsFor: 'private' stamp: ''!
keySize: numberOfBitsInteger

	keySize _ numberOfBitsInteger
! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: ''!
lowestPossibleKey

	^0
! !

!MaMagnitudeIndex methodsFor: 'private' stamp: 'cmm 10/31/2002 19:37'!
open

	self open: self filename! !

!MaMagnitudeIndex methodsFor: 'private' stamp: 'cmm 8/22/2005 09:24'!
open: aReadWriteStream

	super open: aReadWriteStream.
	nextRecordNumber := broker nextRecordNumber.
	nextRecordNumber isInteger
		ifFalse:
			[ MaSpecialCollectionsSoftwareError signal: 'Invalid index size.  Requires maintenance tool.' ]! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: ''!
recordSize

	self subclassResponsibility
! !

!MaMagnitudeIndex methodsFor: 'framework' stamp: 'cmm 12/23/2002 22:16'!
size

	self subclassResponsibility! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: 'cmm 4/28/2005 21:12'!
valueSize

	^ valueSize! !

MaObject subclass: #MaPermutator
	instanceVariableNames: 'arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections'!

!MaPermutator class methodsFor: 'actions' stamp: ''!
permutationsOf: anArrayOfArrays do: aBlock
	"For each combination of elements in anArrayOfArray's, value aBlock
	with the arguments of each combination of elements.  There must be
	as many arguments in aBlock as there are elements in the outer array
	of Array's.

	Note, this also works with any sequenceable collection of sequenceable
	collections."

	| numberOfCombinations args |

	numberOfCombinations :=
		anArrayOfArrays
			inject: 1
			into: [ :prod :each | each size * prod ].
	args := Array new: anArrayOfArrays size.
	1
		to: numberOfCombinations
		do:
			[ :elementIndex |
			1
				to: anArrayOfArrays size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := anArrayOfArrays at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (anArrayOfArrays at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			aBlock valueWithArguments: args ]
! !

!MaPermutator methodsFor: 'building' stamp: ''!
addArguments: aSequenceableCollection

	^self arguments add: aSequenceableCollection
! !

!MaPermutator methodsFor: 'accessing' stamp: ''!
arguments

	^arguments! !

!MaPermutator methodsFor: 'building' stamp: ''!
arguments: aSequenceableCollectionOfSequenceableCollections

	arguments := aSequenceableCollectionOfSequenceableCollections
! !

!MaPermutator methodsFor: 'private initializing' stamp: 'cmm 12/24/2002 15:51'!
initialize
	super initialize.
	arguments := OrderedCollection new! !

!MaPermutator methodsFor: 'actions' stamp: ''!
permutationsDo: oneArgBlock
	"For each combination of elements in the receivers args, value aBlock
	with the arguments of each combination of elements.  The argument to
	oneArgBlock is the Array of parameters for each combination."

	| args |

	args := Array new: self arguments size.
	1
		to: self size
		do:
			[ :elementIndex |
			1
				to: self arguments size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := self arguments at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (self arguments at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			oneArgBlock value: args ]
! !

!MaPermutator methodsFor: 'actions' stamp: ''!
permutationsWhere: conditionBlock do: oneArgBlock
	"Same as permutationsDo: except the conditionBlock must also be
	satisfied for aBlock to run.  This is only for mildly better looking code
	since a case statement could achieve the same effect."

	| args |

	args := Array new: self arguments size.
	1
		to: self size
		do:
			[ :elementIndex |
			1
				to: self arguments size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := self arguments at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (self arguments at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			(conditionBlock value: args) ifTrue: [ oneArgBlock value: args ] ]
! !

!MaPermutator methodsFor: 'accessing' stamp: ''!
size

	^self arguments
		inject: 1
		into: [ :prod :each | each size * prod ]
! !

MaObject subclass: #MaRecordBroker
	instanceVariableNames: 'headerSize recordSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Records'!

!MaRecordBroker commentStamp: 'cmm 5/18/2005 09:46' prior: 0!
I am an abstract superclass for record-based collections such as MaHashIndex and MaLargeArrayOfNumbers.  By plugging them with my appropriate subclass, they are able to be file-based or in-memory (or maybe remote network location in the future).

I don't actually deal with the records themselves, just their reading and writing to the storage.!

MaRecordBroker subclass: #MaFileRecordBroker
	instanceVariableNames: 'file'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Records'!

!MaFileRecordBroker methodsFor: 'file-specific' stamp: 'cmm 5/17/2005 23:32'!
close

	file ifNotNil: [ file closed ifFalse: [ file close ] ]! !

!MaFileRecordBroker methodsFor: 'file-specific' stamp: 'cmm 6/13/2005 23:22'!
create: aFileStream

	file _ aFileStream ifNil: [ ReadWriteStream on: (ByteArray new: 1000) ]! !

!MaFileRecordBroker methodsFor: 'file-specific' stamp: 'cmm 5/18/2005 13:30'!
delete

	self close.
	(FileDirectory forFileName: file name) deleteFileNamed: file name! !

!MaFileRecordBroker methodsFor: 'file-specific' stamp: 'cmm 5/30/2006 08:49'!
filename

	^ file filename! !

!MaFileRecordBroker methodsFor: 'file-specific' stamp: 'cmm 5/18/2005 14:30'!
filePositionFor: recordIndex

	^ recordIndex * recordSize + headerSize! !

!MaFileRecordBroker methodsFor: 'overriding' stamp: 'cmm 5/23/2005 14:11'!
nextRecordNumber

	^ (file size - self headerSize) / recordSize! !

!MaFileRecordBroker methodsFor: 'file-specific' stamp: 'cmm 6/14/2005 10:31'!
open: aReadWriteStream

	file _ aReadWriteStream! !

!MaFileRecordBroker methodsFor: 'file-specific' stamp: 'cmm 5/17/2005 23:28'!
readHeaderInto: aByteArray

	file
		maRead: self headerSize
		bytesFromPosition: 1
		of: aByteArray
		atFilePosition: 0! !

!MaFileRecordBroker methodsFor: 'record' stamp: 'cmm 5/22/2005 22:53'!
readInto: aMaFixedLengthRecord at: recordIndex

	file
		maRead: recordSize
		bytesFromPosition: 1
		of: aMaFixedLengthRecord byteArray
		atFilePosition: (self filePositionFor: recordIndex)! !

!MaFileRecordBroker methodsFor: 'overriding' stamp: 'cmm 5/23/2005 14:12'!
size

	^ self nextRecordNumber! !

!MaFileRecordBroker methodsFor: 'record' stamp: 'cmm 5/22/2005 22:54'!
writeFrom: aMaFixedLengthRecord at: recordIndex

	file
		maWrite: recordSize
		bytesFromPosition: 1
		of: aMaFixedLengthRecord byteArray
		atFilePosition: (self filePositionFor: recordIndex)! !

!MaFileRecordBroker methodsFor: 'file-specific' stamp: 'cmm 7/11/2005 09:03'!
writeHeaderFrom: aByteArray

	file
		maWrite: self headerSize
		bytesFromPosition: 1
		of: aByteArray
		atFilePosition: 0! !

MaRecordBroker subclass: #MaInImageRecordBroker
	instanceVariableNames: 'records'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Records'!

!MaInImageRecordBroker methodsFor: 'overriding' stamp: 'cmm 5/23/2005 23:24'!
close
	"nothing to do.."! !

!MaInImageRecordBroker methodsFor: 'initialize' stamp: 'cmm 5/23/2005 17:22'!
initialize

	super initialize.
	records _ OrderedCollection new! !

!MaInImageRecordBroker methodsFor: 'initialize' stamp: 'cmm 5/23/2005 14:34'!
nextRecordNumber

	^ records size! !

!MaInImageRecordBroker methodsFor: 'overriding' stamp: 'cmm 5/25/2005 16:24'!
readInto: aMaFixedLengthRecord at: recordIndex
	"Since records can be 0-based (sigh), we +1 the recordIndex."

	aMaFixedLengthRecord byteArray: (records at: recordIndex+1) byteArray copy! !

!MaInImageRecordBroker methodsFor: 'overriding' stamp: 'cmm 5/23/2005 23:37'!
writeFrom: aMaFixedLengthRecord at: recordIndex

	(records 
		at: recordIndex+1
		ifAbsentPut: aMaFixedLengthRecord copy) byteArray: aMaFixedLengthRecord byteArray copy! !

!MaRecordBroker class methodsFor: 'as yet unclassified' stamp: 'cmm 5/17/2005 23:52'!
headerSize: anInteger

	^ self new
		headerSize: anInteger ;
		yourself! !

!MaRecordBroker methodsFor: 'override' stamp: 'cmm 5/23/2005 14:32'!
delete
	"Do nothing by default."! !

!MaRecordBroker methodsFor: 'override' stamp: 'cmm 5/17/2005 22:52'!
filename
	self subclassResponsibility ! !

!MaRecordBroker methodsFor: 'accessing' stamp: 'cmm 6/5/2005 12:55'!
fileSize
	"Typically only needed for file-based brokers, but needed here for compatibility of non file-based brokers."

	^ self size * self recordSize! !

!MaRecordBroker methodsFor: 'accessing' stamp: 'cmm 5/17/2005 23:49'!
headerSize

	^ headerSize! !

!MaRecordBroker methodsFor: 'accessing' stamp: 'cmm 5/17/2005 23:49'!
headerSize: anInteger

	headerSize _ anInteger! !

!MaRecordBroker methodsFor: 'override' stamp: 'cmm 5/17/2005 23:01'!
nextRecordNumber

	self subclassResponsibility ! !

!MaRecordBroker methodsFor: 'override' stamp: 'cmm 5/17/2005 23:01'!
open: fileNameString

	self subclassResponsibility ! !

!MaRecordBroker methodsFor: 'override' stamp: 'cmm 5/23/2005 22:50'!
readInto: aMaFixedLengthRecord at: recordIndex

	self subclassResponsibility! !

!MaRecordBroker methodsFor: 'accessing' stamp: 'cmm 5/22/2005 21:15'!
recordSize

	^ recordSize! !

!MaRecordBroker methodsFor: 'accessing' stamp: 'cmm 5/22/2005 21:15'!
recordSize: anInteger

	recordSize _ anInteger! !

!MaRecordBroker methodsFor: 'override' stamp: 'cmm 5/23/2005 23:04'!
writeFrom: aMaFixedLengthRecord at: recordIndex

	self subclassResponsibility! !

MaObject subclass: #MaTree
	instanceVariableNames: 'parents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Tree'!

!MaTree methodsFor: 'building' stamp: ''!
addAll: aCollection

	"Adds all in aCollection as a main parents in the receiver."

	aCollection do: [ :each | self add: each ]
! !

!MaTree methodsFor: 'building' stamp: 'cmm 6/10/2003 22:59'!
addPath: anOrderedCollection

	"Adds each object in anOrderedCollection as a new path in
	the receiver.  The first element in anOrderedCollection is the
	main parent, and each subsequent element is a child of the
	previous.  Any parts of the path that need created will be."

	^self
		addPath: anOrderedCollection
		to: nil
! !

!MaTree methodsFor: 'building' stamp: ''!
addPath: anOrderedCollection to: parentObject
	"Adds each object in anOrderedCollection as a new path in
	the receiver.  The first element in anOrderedCollection is the
	main parent, and each subsequent element is a child of the
	previous.  If part of the path already exists, there it is
	replaced (though they should be equivalent anyway)."

	anOrderedCollection
		inject: parentObject
		into:
			[ :parent :each |
			self
				add: each
				asChildOf: parent.
			each ].
	^anOrderedCollection
! !

!MaTree methodsFor: 'building' stamp: ''!
addTree: aMaTree

	^self addTree: aMaTree to: nil
! !

!MaTree methodsFor: 'building' stamp: 'cmm 12/24/2002 16:08'!
addTree: aMaTree to: anObject

	(self includes: anObject) ifFalse: [ self error: anObject printString, ' is not part of ', self printString ].
	aMaTree withParentsDo: [ :each :eachParent | self add: each asChildOf: eachParent ].
	^aMaTree
! !

!MaTree methodsFor: 'building' stamp: ''!
add: anObject
	"Adds anObject as a main parent in the receiver."

	self
		add: anObject
		asChildOf: nil.
	^anObject
! !

!MaTree methodsFor: 'building' stamp: 'cmm 2/12/2004 00:02'!
add: anObject asChildOf: aParentObject

	parents at: anObject put: aParentObject.
	^anObject! !

!MaTree methodsFor: 'enumerating' stamp: ''!
allLeafsDo: aBlock
	"Value aBlock with each element in the receiver that has no children."

	| children |

	children := self children.
	parents keysAndValuesDo:
		[ :eachChild :eachParent |
		(children includesKey: eachChild) ifFalse: [ aBlock value: eachChild ] ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
allMembers
	"Returns a Set of all the members of the hierarchy"

	^parents keys! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:04'!
allMembersDo: aBlock
	"Returns a Set of all the members of the hierarchy"

	^parents keysDo: aBlock! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 1/23/2002 14:30'!
children

	| children |

	children := Dictionary new: parents size.
	parents keysAndValuesDo:
		[ :eachChild :eachParent |
		(children
			at: eachParent
			ifAbsentPut: [ OrderedCollection new ])
				add: eachChild ].
	^children
! !

!MaTree methodsFor: 'accessing' stamp: ''!
childrenOf: anObject 
	^parents maAllKeysAtValue: anObject! !

!MaTree methodsFor: 'accessing' stamp: ''!
collect: aBlock
	"Returns a new tree whose nodes have all been operated on by aBlock."

	^self
		collect: aBlock
		into: self class new
! !

!MaTree methodsFor: 'private' stamp: 'cmm 1/23/2002 14:30'!
collect: aBlock into: aMaTree 
	"Since some objects may share a parent, we use the map to
	make sure we only value each parent just once, otherwise we
	could have some identity problems."

	| map |
	map := Dictionary new.
	parents keysAndValuesDo: 
			[:eachKey :eachValue | 
			eachValue isNil 
				ifTrue: [aMaTree add: (map at: eachKey ifAbsentPut: [aBlock value: eachKey])]
				ifFalse: 
					[aMaTree add: (map at: eachKey ifAbsentPut: [aBlock value: eachKey])
						asChildOf: (map at: eachValue ifAbsentPut: [aBlock value: eachValue])]].
	^aMaTree! !

!MaTree methodsFor: 'accessing' stamp: ''!
conform: aBlock 
	parents keysDo: [:each | (aBlock value: each) ifFalse: [^false]].
	^true! !

!MaTree methodsFor: 'building' stamp: 'cmm 3/24/2005 11:06'!
copy

	^ super copy
		parents: parents ;
		yourself! !

!MaTree methodsFor: 'enumerating' stamp: ''!
depthFirstDo: twoArgBlock
	"Value twoArgBlock with each object in the receiver and its parent in the order
	of depthFirst."

	self roots do:
		[ :each |
		twoArgBlock
			value: each
			value: nil.
		self
			depthFirstDo: twoArgBlock
			at: each ]
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
depthFirstDo: twoArgBlock at: anObject

	(self childrenOf: anObject) do:
		[ :each |
		twoArgBlock
			value: each
			value: anObject.
		self
			depthFirstDo: twoArgBlock
			at: each ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
depthOf: anObject
	"Returns the number of parents anObject has."

	| depth prevChild |

	(self includes: anObject)
		ifFalse:
			[ self error: self printString , ' does not include ' , anObject printString ].
	depth := 0.
	prevChild := anObject.
	[ prevChild isNil ] whileFalse:
		[ prevChild := self parentOf: prevChild.
		depth := depth + 1 ].
	^depth
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:02'!
detect: aBlock

	^parents detect: aBlock
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:02'!
detect: aBlock ifNone: exceptionBlock

	^parents
		detect: aBlock
		ifNone: exceptionBlock value
! !

!MaTree methodsFor: 'enumerating' stamp: 'cmm 12/24/2002 16:07'!
do: oneArgBlock

	parents keysDo: oneArgBlock! !

!MaTree methodsFor: 'testing' stamp: 'cmm 7/25/2002 19:43'!
hasChildFor: anObject 
	"Returns true if anObject has any children.  Signal an error
	if anObject is not part of the hierarchy."

	^parents includes: anObject! !

!MaTree methodsFor: 'testing' stamp: ''!
hasParentFor: anObject
	"Returns true if anObject has a parent"

	^(self parentOf: anObject) notNil
! !

!MaTree methodsFor: 'testing' stamp: ''!
includes: anObject 
	^parents includesKey: anObject! !

!MaTree methodsFor: 'private' stamp: 'cmm 3/24/2005 11:07'!
initialize
	super initialize.
	self parents: Dictionary new! !

!MaTree methodsFor: 'enumerating' stamp: 'cmm 6/10/2003 23:02'!
maDetect: aBlock ifFound: foundBlock ifNone: exceptionBlock 
	^parents
		maDetect: aBlock
		ifFound: foundBlock
		ifNone: exceptionBlock value! !

!MaTree methodsFor: 'building' stamp: 'cmm 6/10/2003 23:00'!
move: anObject to: newParentObject 
	^parents at: anObject put: newParentObject! !

!MaTree methodsFor: 'accessing' stamp: ''!
nodesInPathFrom: someChildObject to: someParentObject
	"Returns a collection of nodes between someChildObject
	and someParentObject."

	|answer node|

	answer := OrderedCollection new.
	self
		nodesInPathFrom: someChildObject
		to: someParentObject
		do: [ :each | answer add: each ].
	^answer
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
nodesInPathFrom: someChildObject to: someParentObject do: oneArgBlock
	"For each node in the tree from someChildObject to someParentObject, value
	oneArgBlock.  The highest level parent (nil) will not be used as an argument
	to value oneArgBlock."

	| node |

	node := someChildObject.
	oneArgBlock value: someChildObject.
	[ node isNil or: [ node = someParentObject ] ] whileFalse:
		[ oneArgBlock value: node.
		node := self parentOf: node ].
	someParentObject notNil ifTrue: [ oneArgBlock value: someParentObject ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
parentOf: anObject 
	^parents at: anObject! !

!MaTree methodsFor: 'private' stamp: 'cmm 3/24/2005 11:07'!
parents: aDictionary

	parents _ aDictionary! !

!MaTree methodsFor: 'accessing' stamp: ''!
pathFrom: anObject
	"Returns a collection of the objects that are contained in
	the path to anObject ordered from an Object to the top of the hierarchy."

	^self
		nodesInPathFrom: anObject
		to: nil
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
pathFrom: anObject do: aBlock
	"Value aBlock with the objects from anObject to its highest level parent."

	^self
		nodesInPathFrom: anObject
		to: nil
		do: aBlock
! !

!MaTree methodsFor: 'accessing' stamp: ''!
pathTo: anObject
	"Returns a collection of the objects that are contained in
	the path to anObject, ordered from the the top
	of the hierarchy to anObject."

	^(self pathFrom: anObject) reverse
! !

!MaTree methodsFor: 'testing' stamp: ''!
pathTo: anObject includes: anotherObject
	"Returns a boolean.  Example:  in the Smalltalk
	class hierarchy, pathTo: Integer includes: Magnitude would
	return true.  We use pathFrom in the implementation because
	it is faster."

	^(self pathFrom: anObject) includes: anotherObject
! !

!MaTree methodsFor: 'printing' stamp: 'cmm 12/24/2002 16:08'!
printOn: aStream

	super printOn: aStream.
	self depthFirstDo:
		[ :eachChild :eachParent |
		aStream
			cr;
			tab: (self depthOf: eachChild);
			maPrintAbbreviated: eachChild ]
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 7/1/2004 00:34'!
reject: aBlock 
	"Returns a new Tree with all branches removed that
	meet aBlock criteria.  See method comment for more
	info."

	^self reject: [:each | aBlock value: each ]! !

!MaTree methodsFor: 'building' stamp: 'cmm 7/25/2002 19:42'!
remove: anObject
	"Removes anObject and all of its children."

	^self
		remove: anObject
		ifAbsent: [ self error: anObject printString, ' was not found in ', self printString ]
! !

!MaTree methodsFor: 'building' stamp: ''!
remove: anObject ifAbsent: oneArgBlock 
	"Answers the collection of removed objects removed after removing
	anObject after removing it and all of its children."

	| tree |
	(tree := self treeAt: anObject) allMembers 
		do: [:each | parents removeKey: each]! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:08'!
roots
	"Answers a collection of all roots in this tree."

	^self childrenOf: nil
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 4/10/2005 18:00'!
select: aBlock 
	"Returns a new Tree with all branches removed that
	meet aBlock criteria.  See method comment for more
	info."

	| newTree |
	newTree := self maOriginalClass new.
	self allMembersDo: 
			[:each | 
			(aBlock value: each) ifTrue: [newTree addPath: (self pathTo: each)]].
	^newTree! !

!MaTree methodsFor: 'accessing' stamp: ''!
size
	^parents size! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 7/25/2002 19:43'!
treeAt: aNode

	"Return a new instance of the receiver which has a single main
	parent aNode."

	^self
		treeAt: aNode
		ifAbsent: [ self error: aNode printString, ' was not found in ', self printString ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
treeAt: anObject ifAbsent: aBlock
	"Return a new instance of the receiver which has a single main
	parent anObject."

	| newTree |

	(self includes: anObject) ifFalse: [ ^aBlock value ].
	newTree := self class new.
	self
		depthFirstDo:
			[ :each :eachParent |
			newTree
				add: each
				asChildOf: eachParent ]
		at: anObject.
	^newTree
! !

!MaTree methodsFor: 'enumerating' stamp: 'cmm 12/24/2002 16:09'!
withParentsDo: twoArgBlock

	parents keysAndValuesDo: twoArgBlock! !

PackageInfo subclass: #MaSpecialCollectionsPackageInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Info'!

!MaSpecialCollectionsPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 12/29/2005 23:55'!
readmeText

	^'I provide several independent, special-purpose collection classes that seem to be generally useful.

MaWeakIdentityKeyDictionary and MaWeakValueDictionary wrap a collection of their Smalltalk counterparts and keeps them all under 12-bits in size, thus minimizing the number of collisions due to Squeaks 12-bits identity hash.  These provide major improvement in the #removeKey: message.

MaTree is one of my first attempts at a Smalltalk collection.  I never actually used it though I decided to keep it around.  It''s API compatible with the other collections.  To play with an example:

	MaTree new
		add: ''Great Grandma'';
		add: ''Grandma'' asChildOf: ''Great Grandma'';
		add: ''Mom'' asChildOf: ''Grandma'';
		yourself


MaIntervalCollection is can efficiently track interval ranges and merge them when they get within their proximityThreshold.  This is useful for recording occupied space in a memory map, or perhaps grouping entities based on their proximity to each other.  Example:

	(MaIntervalCollection proximityThreshold: 1)
		addFrom: 5 to: 10;
		addFrom: 15 to: 20;
		addFrom: 25 to: 30;
		addFrom: 21 to: 24;
		yourself


MaLargeArrayOfNumbers is an auto-growing, never-shrinking array of numbers.  All numbers have the same number of bits available to represent them.  It maintains the collection in a single randomly-accessed file, and is therefore constrained by available disk space or the addressability of the file.

	MaLargeArrayOfNumbers create: filenameString bitSize: 32.
	myArray _ MaLargeArrayOfNumbers open: filenameString.
	myArray
		at: 10 put: 8734 ;
		at: 11 put: 6199 ;
		yourself
	"be sure to #close it when you are done."


MaHashIndex is another file-based, large-collection of uniformly-sized numbers.  It is much slower but more flexible than MaLargeArrayOfNumbers.  It allows pages between any two key ranges via:

	#upTo: countInteger keysAndValuesFrom: lowKey to: highKey startingAt: index

and pages between any two absolute positions via:

	#upTo: countInteger keysAndValuesFromIndex: lowIndex

Other features are:
	- supports key and value sizes from 16 to 256 bits
	- supports duplicate keys
	- includes test cases that demonstrate its capability and correctness

	MaHashIndex 
		create: fileNameString 
		keySize: 64 
		hashRecordSize: 10.
	myIndex _ MaHashIndex open: filenameString.
	myIndex
		add: 8734 at: 10 ;
		add: 1212 at: 10 ;  "duplicate key ok"
		add: 6199 at: 11 ;
		yourself.
	"be sure to #close it when you are done."
'! !

MaSoftwareError subclass: #MaSpecialCollectionsSoftwareError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-exceptions'!

MaUserError subclass: #MaHashIndexUserError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Exceptions'!
